#!/usr/bin/env python3
#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth import Mux, Cat, C

from ..stream.generator  import StreamSerializer
from ..usb.usb2.transfer import USBInStreamInterface
from ..interface.flash   import ECP5ConfigurationFlashInterface, FlashUIDReader


class ECP5FlashUIDStringDescriptor(StreamSerializer):
    """ Custom runtime string descriptor that returns the Flash UID as a hex string. """

    def __init__(self):
        super().__init__(34, domain="usb", stream_type=USBInStreamInterface, max_length_width=16)

    def elaborate(self, platform):
        m = super().elaborate(platform)

        # Create the required modules to retrieve the flash UID.
        m.submodules.spi_bus  = spi_bus  = ECP5ConfigurationFlashInterface(bus=platform.request('spi_flash'), use_cs=True)
        m.submodules.flashuid = flashuid = FlashUIDReader(bus=spi_bus)

        # Helper function to convert a value to a hex string
        def get_hex_char(value, index):
            nibble = value.word_select(index ^ 1, 4)
            return Mux(nibble > 9, nibble - 10 + ord('a'), Cat(nibble, C(0b11, 2)))

        # Create string descriptor
        m.d.comb += [
            self.data[0].eq(self.data_length),
            self.data[1].eq(3),  # string type
        ]
        for i in range(16):
            m.d.comb += self.data[2+2*i].eq(get_hex_char(flashuid.uid, i))

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

"""
This module contains gateware designed to assist with endpoint/transfer state management.
Its components facilitate data transfer longer than a single packet.
"""

from amaranth            import Signal, Elaboratable, Module, Array
from amaranth.lib.memory import Memory

from .packet             import HandshakeExchangeInterface, TokenDetectorInterface
from ..stream            import USBInStreamInterface
from ...stream           import StreamInterface

class USBInTransferManager(Elaboratable):
    """ Sequencer that converts a long data stream (a USB *transfer*) into a burst of USB packets.

    This module is designed so it can serve as the core of a IN endpoint.

    Attributes
    ----------

    active: Signal(), input
        Held high to enable this module to send packets to the host, and interpret tokens from the host.
        This is typically equivalent to the relevant endpoint being addressed by the host.

    transfer_stream: StreamInterface, input stream
        Input stream; accepts transfer data to be sent on the endpoint. This stream represents
        a USB transfer, and can be as long as is desired; and will be sent in max-packet-size chunks.

        For this stream: ``first`` is ignored; and thus entirely optional. ``last`` is optional;
        if it is not provided; this module will send only max-length-packets, sending a new packet
        every time a full packet size is reached.
    packet_stream: USBInStreamInterface, output stream
        Output stream; broken into packets to be sent.

    flush: Signal(), input
        If high, data that is currently buffered will be sent out as soon as possible. The module will
        not wait further for the input stream to end, or to have a max-length packet; it will send a
        packet as soon as possible.

    discard: Signal(), input
        If high, data that is currently buffered will be discarded. The module will not buffer further
        data, or send any further packets, until this signal is low again.

    data_pid: Signal(2), output
        The LSBs of the data PID to be issued with the current packet. Used with :attr:`packet_stream`
        to indicate the PID of the transmitted packet.

    tokenizer: TokenDetectorInterface, input
        Connection to a detector that detects incoming tokens packets.

    handshakes_in: HandshakeExchangeInterface, input
        Indicates when handshakes are received from the host.
    handshakes_out: HandshakeExchangeInterface, output
        Output that carries handshake packet requests.

    generate_zlps: Signal(), input
        If high, zero-length packets will automatically be generated if the end of a transfer would
        not result in a short packet. (This should be set for control endpoints; and for any interface
        where transfer boundaries are significant.)

    start_with_data1: Signal(), input
        If high, the transmitter will start our PID with DATA1
    reset_sequence: Signal(), input
        If true, our PID generated will reset to the value indicated by `start_with_data1`.
        If desired, this can be held permanently high to control our PID expectation manually.

    Parameters
    ----------
    max_packet_size: int
        The maximum packet size for our associated endpoint, in bytes.
    """

    def __init__(self, max_packet_size):

        self._max_packet_size = max_packet_size

        #
        # I/O port
        #
        self.active           = Signal()

        self.transfer_stream  = StreamInterface()
        self.packet_stream    = USBInStreamInterface()

        self.flush            = Signal()
        self.discard          = Signal()

        # Note: we'll start with DATA1 in our register; as we'll toggle our data PID
        # before we send.
        self.data_pid         = Signal(2, init=1)
        self.buffer_toggle    = Signal()

        self.tokenizer        = TokenDetectorInterface()
        self.handshakes_in    = HandshakeExchangeInterface(is_detector=True)
        self.handshakes_out   = HandshakeExchangeInterface(is_detector=False)

        self.generate_zlps    = Signal()
        self.start_with_data1 = Signal()
        self.reset_sequence   = Signal()


    def elaborate(self, platform):
        m = Module()

        #
        # Transciever state.
        #


        # Handle our PID-sequence reset.
        # Note that we store the _inverse_ of our data PID, as we'll toggle our DATA PID
        # before sending. However, if it has already been toggled then this is overridden below.
        with m.If(self.reset_sequence):
            m.d.usb += self.data_pid.eq(~self.start_with_data1)

        #
        # Transmit buffer.
        #
        # Our USB connection imposed a few requirements on our stream:
        # 1) we must be able to transmit packets at a full rate; i.e.
        #    must be asserted from the start to the end of our transfer; and
        # 2) we must be able to re-transmit data if a given packet is not ACK'd.
        #
        # Accordingly, we'll buffer a full USB packet of data, and then transmit
        # it once either a) our buffer is full, or 2) the transfer ends (last=1).
        #
        # This implementation is double buffered; so a buffer fill can be pipelined
        # with a transmit.
        #

        # We'll create two buffers; so we can fill one as we empty the other.
        m.submodules.transmit_buffer_0 = transmit_buffer_0 = Memory(shape=8, depth=self._max_packet_size, init=[])
        m.submodules.transmit_buffer_1 = transmit_buffer_1 = Memory(shape=8, depth=self._max_packet_size, init=[])
        buffer = Array([transmit_buffer_0, transmit_buffer_1])
        buffer_write_ports = Array(buffer[i].write_port(domain="usb") for i in range(2))
        buffer_read_ports  = Array(buffer[i].read_port(domain="usb") for i in range(2))

        # Create values equivalent to the buffer numbers for our read and write buffer; which switch
        # whenever we swap our two buffers.
        write_buffer_number =  self.buffer_toggle
        read_buffer_number  = ~self.buffer_toggle

        # Create a shorthand that refers to the buffer to be filled; and the buffer to send from.
        # We'll call these the Read and Write buffers.
        buffer_write = buffer_write_ports[write_buffer_number]
        buffer_read  = buffer_read_ports[read_buffer_number]

        # Buffer state tracking:
        # - Our ``fill_count`` keeps track of how much data is stored in a given buffer.
        # - Our ``stream_ended`` bit keeps track of whether the stream ended while filling up
        #   the given buffer. This indicates that the buffer cannot be filled further; and, when
        #   ``generate_zlps`` is enabled, is used to determine if the given buffer should end in
        #   a short packet; which determines whether ZLPs are emitted.
        buffer_fill_count   = Array(Signal(range(0, self._max_packet_size + 1)) for _ in range(2))
        buffer_stream_ended = Array(Signal(name=f"stream_ended_in_buffer{i}") for i in range(2))

        # Create shortcuts to active fill_count / stream_ended signals for the buffer being written.
        write_fill_count   = buffer_fill_count[write_buffer_number]
        write_stream_ended = buffer_stream_ended[write_buffer_number]

        # Create shortcuts to the fill_count / stream_ended signals for the packet being sent.
        read_fill_count   = buffer_fill_count[read_buffer_number]
        read_stream_ended = buffer_stream_ended[read_buffer_number]

        # Keep track of our current send position; which determines where we are in the packet.
        send_position = Signal(range(0, self._max_packet_size + 1))

        # Shortcut names.
        in_stream  = self.transfer_stream
        out_stream = self.packet_stream


        # Use our memory's two ports to capture data from our transfer stream; and two emit packets
        # into our packet stream. Since we'll never receive to anywhere else, or transmit to anywhere else,
        # we can just unconditionally connect these.
        m.d.comb += [

            # We'll only ever -write- data from our input stream...
            buffer_write_ports[0].data   .eq(in_stream.payload),
            buffer_write_ports[0].addr   .eq(write_fill_count),
            buffer_write_ports[1].data   .eq(in_stream.payload),
            buffer_write_ports[1].addr   .eq(write_fill_count),

            # ... and we'll only ever -send- data from the Read buffer.
            buffer_read.addr             .eq(send_position),
            out_stream.payload           .eq(buffer_read.data),

            # We're ready to receive data iff we have space in the buffer we're currently filling.
            in_stream.ready              .eq((write_fill_count != self._max_packet_size) & ~write_stream_ended),
            buffer_write.en              .eq(in_stream.valid & in_stream.ready)
        ]

        # Set both fill counts to zero when we discard data.
        with m.If(self.discard):
            m.d.usb += [
                write_fill_count    .eq(0),
                write_stream_ended  .eq(0),
                read_fill_count     .eq(0),
                read_stream_ended   .eq(0),
            ]

        # Increment our fill count whenever we accept new data.
        with m.Elif(buffer_write.en):
            m.d.usb += write_fill_count.eq(write_fill_count + 1)

        # If the stream ends while we're adding data to the buffer, mark this as an ended stream.
        with m.If(in_stream.last & buffer_write.en):
            m.d.usb += write_stream_ended.eq(1)

        # A packet is completing when:
        # - There is a byte arriving from the input stream, and:
        # - It is either the last byte from the stream, or will cause us to reach max packet size.
        packet_nearly_full = (write_fill_count + 1 == self._max_packet_size)
        packet_completing = in_stream.valid & (in_stream.last | packet_nearly_full)

        # We should also send a packet when both:
        # - Our flush input is asserted
        # - We have some data in the buffer to flush.
        packet_to_flush = self.flush & (write_fill_count != 0)

        # We're ready to send a packet when either of the above conditions is met, and not discarding.
        packet_ready = (packet_completing | packet_to_flush) & ~self.discard

        # Shortcut for when we need to deal with an in token.
        # Pulses high an interpacket delay after receiving an IN token.
        in_token_received = self.active & self.tokenizer.is_in & self.tokenizer.ready_for_response

        with m.FSM(domain='usb'):

            # WAIT_FOR_DATA -- We don't yet have a full packet to transmit, so  we'll capture data
            # to fill the our buffer. At full throughput, this state will never be reached after
            # the initial post-reset fill.
            with m.State("WAIT_FOR_DATA"):

                # We can't yet send data; so NAK any packet requests.
                m.d.comb += self.handshakes_out.nak.eq(in_token_received)

                # If we've just finished a packet, we now have data we can send!
                with m.If(packet_ready):
                    m.next = "WAIT_TO_SEND"
                    m.d.usb += [

                        # We're now ready to take the data we've captured and _transmit_ it.
                        # We'll swap our read and write buffers, and toggle our data PID.
                        self.buffer_toggle  .eq(~self.buffer_toggle),
                        self.data_pid[0]    .eq(~self.data_pid[0]),

                        # Mark our current stream as no longer having ended.
                        read_stream_ended  .eq(0)
                    ]


            # WAIT_TO_SEND -- we now have at least a buffer full of data to send; we'll
            # need to wait for an IN token to send it.
            with m.State("WAIT_TO_SEND"):
                m.d.usb += send_position .eq(0),

                # If discarding data, go back to waiting for new data.
                with m.If(self.discard):
                    # Undo the data PID toggle.
                    m.d.usb += self.data_pid[0].eq(~self.data_pid[0]),
                    m.next = "WAIT_FOR_DATA"

                # If we get a clear halt request while in this state, reset to initial PID.
                with m.Elif(self.reset_sequence):
                    m.d.usb += self.data_pid.eq(self.start_with_data1)

                # Otherwise, once we get an IN token, move to sending a packet.
                with m.Elif(in_token_received):

                    # If we have a packet to send, send it.
                    with m.If(read_fill_count):
                        m.next = "SEND_PACKET"
                        m.d.usb += out_stream.first  .eq(1)

                    # Otherwise, we entered a transmit path without any data in the buffer.
                    with m.Else():
                        m.d.comb += [
                            # Send a ZLP...
                            out_stream.valid  .eq(1),
                            out_stream.last   .eq(1),
                        ]
                        # ... and clear the need to follow up with one, since we've just sent a short packet.
                        m.d.usb += read_stream_ended.eq(0)
                        m.next = "WAIT_FOR_ACK"


            with m.State("SEND_PACKET"):
                last_packet = (send_position + 1 == read_fill_count)

                m.d.comb += [
                    # We're always going to be sending valid data, since data is always
                    # available from our memory.
                    out_stream.valid  .eq(1),

                    # Let our transmitter know when we've reached our last packet.
                    out_stream.last   .eq(last_packet)
                ]

                # Once our transmitter accepts our data...
                with m.If(out_stream.ready):

                    m.d.usb += [
                        # ... move to the next byte in our packet ...
                        send_position     .eq(send_position + 1),

                        # ... and mark our packet as no longer the first.
                        out_stream.first  .eq(0)
                    ]

                    # Move our memory pointer to its next position.
                    m.d.comb += buffer_read.addr  .eq(send_position + 1),

                    # If we've just sent our last packet, we're now ready to wait for a
                    # response from our host.
                    with m.If(last_packet):
                        m.next = 'WAIT_FOR_ACK'


            # WAIT_FOR_ACK -- We've just sent a packet; but don't know if the host has
            # received it correctly. We'll wait to see if the host ACKs.
            with m.State("WAIT_FOR_ACK"):

                # If discarding data, go back to waiting for new data.
                with m.If(self.discard):
                    m.next = "WAIT_FOR_DATA"

                # If the host does ACK...
                with m.Elif(self.handshakes_in.ack):
                    # ... clear the data we've sent from our buffer.
                    m.d.usb += read_fill_count.eq(0)

                    # Figure out if we'll need to follow up with a ZLP. If we have ZLP generation enabled,
                    # we'll make sure we end on a short packet. If this is max-packet-size packet _and_ our
                    # transfer ended with this packet; we'll need to inject a ZLP.
                    follow_up_with_zlp = \
                        self.generate_zlps & (read_fill_count == self._max_packet_size) & read_stream_ended

                    # If we're following up with a ZLP, move back to our "wait to send" state.
                    # Since we've now cleared our fill count; this next go-around will emit a ZLP.
                    with m.If(follow_up_with_zlp):
                        m.d.usb += self.data_pid[0].eq(~self.data_pid[0]),
                        m.next = "WAIT_TO_SEND"

                    # Otherwise, there's a possibility we already have a packet-worth of data waiting
                    # for us in our "write buffer", which we've been filling in the background.
                    # If this is the case, we'll flip which buffer we're working with, toggle our data pid,
                    # and then ready ourselves for transmit.
                    with m.Elif(~in_stream.ready | packet_ready):
                        m.next = "WAIT_TO_SEND"
                        m.d.usb += [
                            self.buffer_toggle .eq(~self.buffer_toggle),
                            self.data_pid[0]   .eq(~self.data_pid[0]),
                            read_stream_ended  .eq(0)
                        ]

                    # If neither of the above conditions are true; we now don't have enough data to send.
                    # We'll wait for enough data to transmit.
                    with m.Else():
                        m.next = "WAIT_FOR_DATA"


                # If the host starts a new packet without ACK'ing, we'll need to retransmit, unless discarding.
                # We'll move back to our "wait for token" state without clearing our buffer.
                with m.If(self.tokenizer.new_token & ~self.discard):
                    m.next = 'WAIT_TO_SEND'

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" U0 link-maintenance timers. """

from amaranth import *


class LinkMaintenanceTimers(Elaboratable):
    """ Timers which ensure link integrity is maintained in U0.

    These timers ensure that we provide enough traffic to maintain link state,
    and move to link recovery if we ever fail to see constant traffic.

    Our two main rules [USB3.2r1: 7.5.6.1]:
        - If we don't send a link command for 10uS, we'll need to issue a
          keepalive packet in order to keep our link stably in U0.
        - If we don't receive a link command for 1mS, we know that the link
          is no longer in good condition, since the other side should have
          been sending keepalives to prevent this long of an idle. We'll have
          to perform link recovery.


    Attributes
    ----------
    link_command_received: Signal(), input
        Strobe that should be asserted when a link command is received.
    link_command_transmitted: Signal(), input
        Strobe that should be asserted when a link command is transmitted.

    schedule_keepalive: Signal(), output
        Strobe that indicates that we'll need to send a keepalive packet.
    transition_to_recovery: Signal(), output
        Strobe that indicates that our link is no longer stable; and we'll
        need to perform link recovery.

    Parameters
    ----------
    ss_clock_frequency: float
        The frequency of our ``ss`` domain clock, in Hz.
    """

    KEEPALIVE_TIMEOUT = 10e-6
    RECOVERY_TIMEOUT  = 1e-3


    def __init__(self, *, ss_clock_frequency=125e6):
        self._clock_frequency = ss_clock_frequency

        #
        # I/O port.
        #
        self.enable                   = Signal()

        self.link_command_received    = Signal()
        self.packet_received          = Signal()

        self.link_command_transmitted = Signal()

        self.schedule_keepalive       = Signal()
        self.transition_to_recovery   = Signal()


    def elaborate(self, platform):
        m = Module()

        # Note that we don't care about rollover on any of our timers; as it's harmless in
        # both cases. For our keepalive, we'll immediately send a link command, which should
        # clear our timer. For our recovery timer, we'll enter recovery and reset this whole
        # thing anyway. :)


        #
        # Keepalive Timer
        #
        keepalive_timeout_cycles = int(self.KEEPALIVE_TIMEOUT * self._clock_frequency)

        # Time how long it's been since we've sent our last link command.
        keepalive_timer = Signal(range(keepalive_timeout_cycles))
        m.d.comb += self.schedule_keepalive.eq(keepalive_timer + 1 == keepalive_timeout_cycles)

        with m.If(self.link_command_transmitted):
            m.d.ss += keepalive_timer.eq(0)
        with m.Elif(self.enable):
            m.d.ss += keepalive_timer.eq(keepalive_timer + 1)
        with m.Else():
            m.d.ss += keepalive_timer.eq(0)


        #
        # Recovery Timer
        #
        recovery_timeout_cycles = int(self.RECOVERY_TIMEOUT * self._clock_frequency)

        # Time how long it's been since we've received our last link command.
        recovery_timer = Signal(range(recovery_timeout_cycles))
        m.d.comb += self.transition_to_recovery.eq(recovery_timer + 1 == recovery_timeout_cycles)

        with m.If(self.link_command_received | self.packet_received):
            m.d.ss += recovery_timer.eq(0)
        with m.Elif(self.enable):
            m.d.ss += recovery_timer.eq(recovery_timer + 1)
        with m.Else():
            m.d.ss += recovery_timer.eq(0)


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Utilities for creating LUNA platforms. """

import logging

from amaranth import Signal, Record
from amaranth.build.res import ResourceError, Subsignal, Resource, Pins


class NullPin(Record):
    """ Stand-in for a I/O record. """

    def __init__(self, size=1):
        super().__init__([
            ('i', size),
            ('o', size),
            ('oe', 1),
        ])


class LUNAPlatform:
    """ Mixin that extends Amaranth platforms with extra functionality."""

    name = "unnamed platform"

    def create_usb3_phy(self):
        """ Shortcut that creates a USB3 phy configured for the given platform, if possible. """

        # Ensure our platform has what it needs to create our USB3 PHY.
        if not hasattr(self, "default_usb3_phy"):
            raise ValueError(f"Platform {self.name} has no default USB3 PHY; cannot create one automatically.")

        # Create our PHY, allowing it access to our object.
        return self.default_usb3_phy(self)


    def get_led(self, m, index=0):
        """ Attempts to get an LED for the given platform, where possible.

        If no LED is available, returns a NullPin, so the design can still be generated
        without the relevant LED.
        """

        # First, try to get a simple LED.
        try:
            return self.request("led", index)
        except ResourceError:
            pass

        # Next, try to get an RGB LED, if the platform has one.
        # If we find one, we'll grab only one leg of it, and turn the others off.
        try:
            rgb_led = self.request("rgb_led", index)
            m.d.comb += [
                rgb_led.r.o.eq(0),
                rgb_led.b.o.eq(0)
            ]
            return rgb_led.g
        except ResourceError:
            pass


        # Finally, if we've failed to get an LED entirely, return a NullPin stand-in.
        return NullPin()


    def request_optional(self, name, number=0, *args, default, expected=False, **kwargs):
        """ Specialized version of .request() for "optional" I/O.

        If the platform has the a resource with the given name, it is requested
        and returned. Otherwise, this method returns the value provided in the default argument.

        This is useful for designs that support multiple platforms; and allows for
        resources such as e.g. LEDs to be omitted on platforms that lack them.

        Parameters
        ----------
        default: any
            The value that is returned in lieu of the relevant resources if the resource does not exist.
        expected: bool, optional
            If explicitly set to True, this function will emit a warning when the given pin is not present.
        """

        try:
            return self.request(name, number, *args, **kwargs)
        except ResourceError:
            log = logging.warnings if expected else logging.debug
            log(f"Skipping resource {name}/{number}, as it is not present on this platform.")
            return default


class LUNAApolloPlatform(LUNAPlatform):
    """ Base class for Apollo-based LUNA platforms; includes configuration. """

    def port_sharing(self, phy_name):
        """ Reports whether and how the USB port for a given PHY is shared with Apollo.

        Parameters
        ----------
        phy_name: str
            The name of a PHY resource on the platform.

        Returns
        -------
        A string identifying the sharing mechanism, or None if the port is not shared.

        Supported sharing mechanisms are:

        "advertising":
            The gateware must create an ApolloAdvertiser instance, which will
            signal to the Apollo MCU that it wishes to use the port. The FPGA
            may hand the port back to Apollo by ceasing advertising.
        """
        sharing = getattr(self, "apollo_port_sharing", {})
        return sharing.get(phy_name, None)


    @property
    def apollo_gateware_phy(self):
        """ Returns the USB PHY for Apollo gateware. """
        
        sharing = getattr(self, "apollo_port_sharing", {})
        if len(sharing) == 0:
            return self.default_usb_connection
        return list(sharing.keys())[0]


    def toolchain_program(self, products, name):
        """ Programs the relevant LUNA board via its sideband connection. """

        from apollo_fpga import ApolloDebugger
        from apollo_fpga.ecp5 import ECP5_JTAGProgrammer

        # Create our connection to the debug module.
        debugger = ApolloDebugger()

        # Grab our generated bitstream, and upload it to the FPGA.
        bitstream =  products.get("{}.bit".format(name))
        with debugger.jtag as jtag:
            programmer = ECP5_JTAGProgrammer(jtag)
            programmer.configure(bitstream)


    def _ensure_unconfigured(self, debugger):
        """ Ensures a given FPGA is unconfigured and thus ready for e.g. SPI flashing. """

        from apollo_fpga.ecp5 import ECP5_JTAGProgrammer

        debugger.force_fpga_offline()


    def toolchain_flash(self, products, name="top"):
        """ Programs the LUNA board's flash via its sideband connection. """

        from apollo_fpga import ApolloDebugger
        from apollo_fpga.ecp5 import ECP5_JTAGProgrammer

        # Create our connection to the debug module.
        debugger = ApolloDebugger()
        self._ensure_unconfigured(debugger)

        # Grab our generated bitstream, and upload it to the .
        bitstream =  products.get("{}.bit".format(name))
        with debugger.jtag as jtag:
            programmer = ECP5_JTAGProgrammer(jtag)
            programmer.flash(bitstream)

        debugger.soft_reset()


    def toolchain_erase(self):
        """ Erases the LUNA board's flash. """

        from apollo_fpga import ApolloDebugger
        from apollo_fpga.ecp5 import ECP5_JTAGProgrammer

        # Create our connection to the debug module.
        debugger = ApolloDebugger()
        self._ensure_unconfigured(debugger)

        with debugger.jtag as jtag:
            programmer = ECP5_JTAGProgrammer(jtag)
            programmer.erase_flash()

        debugger.soft_reset()



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Boilerplate for LUNA unit tests. """

import os
import math
import unittest

from functools import wraps

from amaranth import Signal
from amaranth.sim import Simulator


def sync_test_case(process_function, *, domain="sync"):
    """ Decorator that converts a function into a simple synchronous-process test case. """

    #
    # This function should automatically transform a given function into a pysim
    # synch process _without_ losing the function's binding on self. Accordingly,
    # we'll create a wrapper function that has self bound, and then a test case
    # that's closed over that wrapper function's context.
    #
    # This ensure that self is still accessible from the decorated function.
    #

    def run_test(self):
        @wraps(process_function)
        def test_case():
            yield from self.initialize_signals()
            yield from process_function(self)

        self.domain = domain
        self._ensure_clocks_present()
        self.sim.add_sync_process(test_case, domain=domain)
        self.simulate(vcd_suffix=process_function.__name__)

    return run_test


def usb_domain_test_case(process_function):
    """
    Decorator that converts a function into a simple synchronous-process
    test case in the USB domain.
    """
    return sync_test_case(process_function, domain='usb')


def fast_domain_test_case(process_function):
    """
    Decorator that converts a function into a simple synchronous-process
    test case in the fast domain.
    """
    return sync_test_case(process_function, domain='fast')


def ss_domain_test_case(process_function):
    """
    Decorator that converts a function into a simple synchronous-process
    test case in the SuperSpeed domain.
    """
    return sync_test_case(process_function, domain='ss')



class LunaGatewareTestCase(unittest.TestCase):

    domain = 'sync'

    # Convenience property: if set, instantiate_dut will automatically create
    # the relevant fragment with FRAGMENT_ARGUMENTS.
    FRAGMENT_UNDER_TEST = None
    FRAGMENT_ARGUMENTS = {}

    # Convenience properties: if not None, a clock with the relevant frequency
    # will automatically be added.
    FAST_CLOCK_FREQUENCY = None
    SYNC_CLOCK_FREQUENCY = 120e6
    USB_CLOCK_FREQUENCY  = None
    SS_CLOCK_FREQUENCY   = None


    def instantiate_dut(self):
        """ Basic-most function to instantiate a device-under-test.

        By default, instantiates FRAGMENT_UNDER_TEST.
        """
        return self.FRAGMENT_UNDER_TEST(**self.FRAGMENT_ARGUMENTS)


    def get_vcd_name(self):
        """ Return the name to use for any VCDs generated by this class. """
        return "test_{}".format(self.__class__.__name__)


    def setUp(self):
        self.dut = self.instantiate_dut()
        self.sim = Simulator(self.dut)

        if self.USB_CLOCK_FREQUENCY:
            self.sim.add_clock(1 / self.USB_CLOCK_FREQUENCY, domain="usb")
        if self.SYNC_CLOCK_FREQUENCY:
            self.sim.add_clock(1 / self.SYNC_CLOCK_FREQUENCY, domain="sync")
        if self.FAST_CLOCK_FREQUENCY:
            self.sim.add_clock(1 / self.FAST_CLOCK_FREQUENCY, domain="fast")
        if self.SS_CLOCK_FREQUENCY:
            self.sim.add_clock(1 / self.SS_CLOCK_FREQUENCY, domain="ss")


    def initialize_signals(self):
        """ Provide an opportunity for the test apparatus to initialize siganls. """
        yield Signal()


    def traces_of_interest(self):
        """ Returns an interable of traces to include in any generated output. """
        return ()


    def simulate(self, *, vcd_suffix=None):
        """ Runs our core simulation. """

        # If we're generating VCDs, run the test under a VCD writer.
        if os.getenv('GENERATE_VCDS', default=False):

            # Figure out the name of our VCD files...
            vcd_name = self.get_vcd_name()
            if vcd_suffix:
                vcd_name = "{}_{}".format(vcd_name, vcd_suffix)

            # ... and run the simulation while writing them.
            traces = self.traces_of_interest()
            with self.sim.write_vcd(vcd_name + ".vcd", vcd_name + ".gtkw", traces=traces):
                self.sim.run()

        else:
            self.sim.run()


    @staticmethod
    def pulse(signal, *, step_after=True):
        """ Helper method that asserts a signal for a cycle. """
        yield signal.eq(1)
        yield
        yield signal.eq(0)

        if step_after:
            yield


    @staticmethod
    def advance_cycles(cycles):
        """ Helper methods that waits for a given number of cycles. """

        for _ in range(cycles):
            yield


    @staticmethod
    def wait_until(strobe, *, timeout=None):
        """ Helper method that advances time until a strobe signal becomes true. """

        cycles_passed = 0

        while not (yield strobe):
            yield

            cycles_passed += 1
            if timeout and cycles_passed > timeout:
                raise RuntimeError(f"Timeout waiting for '{strobe.name}' to go high!")


    def _ensure_clocks_present(self):
        """ Function that validates that a clock is present for our simulation domain. """
        frequencies = {
            'sync': self.SYNC_CLOCK_FREQUENCY,
            'usb':  self.USB_CLOCK_FREQUENCY,
            'fast': self.FAST_CLOCK_FREQUENCY,
            'ss': self.SS_CLOCK_FREQUENCY
        }
        self.assertIsNotNone(frequencies[self.domain], f"no frequency provied for `{self.domain}`-domain clock!")


    def wait(self, time):
        """ Helper method that waits for a given number of seconds in a *_test_case. """

        # Figure out the period of the clock we want to work with...
        if self.domain == 'sync':
            period = 1 / self.SYNC_CLOCK_FREQUENCY
        elif self.domain == 'usb':
            period = 1 / self.USB_CLOCK_FREQUENCY
        elif self.domain == 'fast':
            period = 1 / self.FAST_CLOCK_FREQUENCY

        # ... and, accordingly, how many cycles we want to delay.
        cycles = math.ceil(time / period)
        print(cycles)

        # Finally, wait that many cycles.
        yield from self.advance_cycles(cycles)


class LunaUSBGatewareTestCase(LunaGatewareTestCase):
    """ Specialized form of :class:``LunaGatewareTestCase`` that assumes a USB domain clock, but no others. """

    SYNC_CLOCK_FREQUENCY = None
    USB_CLOCK_FREQUENCY  = 60e6


class LunaSSGatewareTestCase(LunaGatewareTestCase):
    """ Specialized form of :class:``LunaGatewareTestCase`` that assumes a USB domain clock, but no others. """

    SYNC_CLOCK_FREQUENCY = None
    SS_CLOCK_FREQUENCY   = 125e6



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code adapted from ``litex`` and ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" SerDes-based USB3 PIPE PHY. """

#
# Quick-use aliases
#
__all__ = ['ECP5SerDesPIPE', 'XC7GTPSerDesPIPE', 'XC7GTXSerDesPIPE']

from .ecp5    import ECP5SerDesPIPE
from .xc7_gtp import XC7GTPSerDesPIPE
from .xc7_gtx import XC7GTXSerDesPIPE



# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 PIPE interfacing gateware. """

from amaranth import *
from amaranth.lib.cdc  import FFSynchronizer, ResetSynchronizer
from amaranth.lib.fifo import AsyncFIFOBuffered


class PIPEInterface(Elaboratable):
    """ Interface present on hardware that implements the PHY Interface for PCI Express and USB 3.0 (PIPE).

    This interface is compliant with the PHY Interface For the PCI Express and USB 3.0
    Architectures, Version 3.0 specification. Unless otherwise noted, the descriptions of
    the signals as stated in this specification take precedence over the ones provided here.

    The directions of the signals are given from the PHY perspective, i.e. a signal described as
    an "output" is driven by the PHY and received by the MAC.

    Parameters
    ----------
    width : int
        Interface width, in symbols.

    Ports
    -----
    reset : Signal(), input
        Active-high, asynchronous reset for the PHY receiver and transmitter.
    clk : Signal(), input
        Reference clock for the PHY receiver and transmitter. The specifications for this clock
        (frequency, jitter, phase relationship to other clocks) are PHY implementation dependent
        and must be specified by the implementation.
    pclk : Signal(), output
        Clock for the PHY interface. Unless otherwise specified, all other signals are synchronous
        to the rising edge of this clock. This clock is derived from the reference clock; if
        the reference clock is spread spectrum modulated, so will be the interface clock.
        The frequency of this clock depends on the ``phy_mode`` and ``rate`` inputs. Note that
        the supported combinations of ``phy_mode`` and ``rate`` are PHY implementation dependent.

    tx_data : Signal(width * 8), input
    tx_datak : Signal(width), input
        Transmit data bus. Bits [7:0] are the first symbol to be transmitted, bits [15:8] (if any)
        are the second symbol, and so on.
    rx_data : Signal(width * 8), output
    rx_datak : Signal(width), output
        Receive data bus. Bits [7:0] are the first symbol that has been received, bits [15:8]
        (if any) are the second symbol, and so on.

    phy_mode : Signal(2), input
        PHY operating mode; 0 for PCI Express, 1 for SuperSpeed USB. The allowed values for this
        input are PHY implementation dependent. The PHY must be reset after changing this input.
    elas_buf_mode : Signal(1), input
        Elastic buffer operating mode; 0 for Nominal Half-full, 1 for Nominal Empty. The allowed
        values for this input are PHY implementation dependent. The PHY must be reset after
        changing this input.
    rate : Signal(1), input
        Link signaling rate; 0 for 2.5 GT/s, 1 for 5 GT/s. The allowed values for this input are
        PHY implementation dependent. Whether changing this signal affects the ``pclk`` frequency
        or the effective data bus width is PHY implementation dependent. The change of this input
        is acknowledged by a single-cycle assertion of ``phy_status``.
    power_down : Signal(2), input
        Protocol-specific power management mode, as per the PIPE specification. This signal is
        synchronous to ``pclk`` if the clock is running, and asynchronous otherwise. The change
        of this input is acknowledged by a single-cycle assertion of ``phy_status`` if ``pclk``
        is running, or by deassertion of ``phy_status`` otherwise.
    tx_deemph : Signal(2), input
        Transmitter de-emphasis level; 0 for -6 dB, 1 for -3.5 dB, 2 for 0 dB. The allowed values
        for this input are PHY implementation dependent.
    tx_margin : Signal(3), input
        Transmitter voltage levels, as per the PIPE specification. The allowed values for this
        input and their meanings are PHY implementation dependent.
    tx_swing : Signal(1)
        Transmitter voltage swing level. The allowed values for this input and their meanings
        are PHY implementation dependent.
    tx_detrx_lpbk : Signal(), input
    tx_elec_idle : Signal(), input
        Protocol-specific transmit control signals, as per the PIPE specification. Depending on
        the state of ``phy_mode`` and ``power_down`` inputs, these inputs direct the PHY
        transmitter to transmit data from the transmit data bus, loop back received data, go into
        Electrical Idle, transmit beacon or LFPS signaling, or perform a receiver detection
        operation. These signals are synchronous to ``pclk`` if the clock is running, and
        asynchronous otherwise. The completion of a receiver detection operation is acknowledged
        by a single-cycle assertion of ``phy_status`` if ``pclk`` is running, or by asserting
        ``phy_status`` until the MAC deasserts ``tx_detrx_lpbk`` otherwise.
    tx_compliance : Signal(), input
        If asserted, sets the running disparity to negative for the first symbol on the transmit
        data bus. This signal is implemented only for PHYs that can operate in the PCI Express mode.
    tx_ones_zeroes : Signal(), input
        If asserted, the PHY transmits an alternating sequence of 50-250 ones and 50-250 zeroes
        instead of the data on the transmit data bus. This signal is implemented only for PHYs
        that can operate in the SuperSpeed USB mode.
    rx_polarity : Signal(), input
        If asserted, the PHY receiver inverts the received serial data.
    rx_eq_training : Signal(), input
        If asserted, the PHY receiver bypasses normal operation to perform equalization training.
        Whether this signal is implemented is PHY implementation dependent.
    rx_termination : Signal(), input
        If asserted, the PHY receiver presents receiver terminations. Whether this signal is
        implemented is PHY implementation dependent.

    phy_status : Signal(), output
        PHY operation completion status, as per the PIPE specification. This signal is synchronous
        to ``pclk`` if the clock is running, and asynchronous otherwise.
    rx_valid : Signal(), output
        If asserted, the PHY receiver has symbol lock and there is valid data on the data bus.
    rx_status : Signal(3), output
        Indicates one of the four possible receiver errors (8b10b decode error, disparity error,
        elastic buffer overflow or underflow), the addition or removal of symbols as a part of
        elastic buffer management, or the completion of a receiver detection operation. This signal
        is synchronous to ``pclk`` if the clock is running, and asynchronous otherwise.
    rx_elec_idle : Signal(), output
        If asserted, depending on the state of ``phy_mode`` and ``power_down`` inputs, indicates
        detection of Electrical Idle, beacon signaling, or LFPS signaling. This signal is
        asynchronous.
    power_present : Signal(), output
        If asserted, voltage is present on Vbus. Whether this signal is implemented is PHY
        implementation dependent.
    """

    # Mappings of interface widths to DataBusWidth parameters.
    _DATA_BUS_WIDTHS = {
        4: 0b00,
        2: 0b01,
        1: 0b10
    }

    def __init__(self, *, width):
        # Ensure we have a valid interface width.
        if width not in (1, 2, 4):
            raise ValueError(f"PIPE does not support a data bus width of {width}")
        self.width          = width

        #
        # Clock and reset signals.
        #
        self.reset          = Signal()
        self.clk            = Signal()
        self.pclk           = Signal()

        #
        # Transmit and receive data buses.
        #
        self.tx_data        = Signal(self.width * 8)
        self.tx_datak       = Signal(self.width * 1)
        self.rx_data        = Signal(self.width * 8)
        self.rx_datak       = Signal(self.width * 1)

        #
        # Control signals.
        #
        self.phy_mode       = Signal(2)
        self.elas_buf_mode  = Signal(1)
        self.rate           = Signal(1)
        self.power_down     = Signal(2)
        self.tx_deemph      = Signal(2)
        self.tx_margin      = Signal(3)
        self.tx_swing       = Signal(1)
        self.tx_detrx_lpbk  = Signal()
        self.tx_elec_idle   = Signal()
        self.tx_compliance  = Signal()
        self.tx_ones_zeroes = Signal()
        self.rx_polarity    = Signal()
        self.rx_eq_training = Signal()
        self.rx_termination = Signal()

        #
        # Status signals.
        #
        self.data_bus_width = Const(self._DATA_BUS_WIDTHS[self.width], 2)
        self.phy_status     = Signal()
        self.rx_valid       = Signal()
        self.rx_status      = Signal(3)
        self.rx_elec_idle   = Signal()
        self.power_present  = Signal()



class AsyncPIPEInterface(PIPEInterface, Elaboratable):
    """ Gateware that transfers PIPE interface signals between clock domains.

    The PIPE specification defines the PHY interface signals to be synchronous to a PHY-generated
    clock ``pclk``, and asynchronous if ``pclk`` is not running. The MAC will typically not be
    clocked by ``pclk`` for the following non-exhaustive list of reasons:
        * The MAC implements P2 (for PCI Express) or P3 (for SuperSpeed USB) power states where
          the PHY-generated clock is not running;
        * The PHY-generated clock is faster than the maximum frequency at which the MAC can run;
        * The MAC implements rate switching (for PCI Express), and needs to support PHYs that
          change the PHY-generated clock frequency depending on the rate;
        * The PHY is required to use a spread spectrum clock in the SuperSpeed USB mode, and
          this would interfere with the MAC operation;
        * etc.

    This gateware transfers the PIPE interface signals between the PHY and MAC clock domains,
    optionally performing gearing to adapt the PHY data bus width and interface clock rate to
    the MAC capabilities. With the exception of ``reset``, ``clk`` and ``pclk``, all of the signals
    in this gateware are synchronous to the specified Amaranth clock domain, ``ss`` by default.
    The ``pclk`` signal is driven by the clock of this domain.

    This gateware does not currently support asynchronous signaling in the deepest PHY power state.
    """

    def __init__(self, phy, *, width, domain="ss"):
        if width < phy.width:
            raise ValueError(f"Async PIPE interface cannot adapt PHY data bus width {phy.width} "
                             f"to MAC data bus width {width}")
        super().__init__(width=width)
        self.phy            = phy
        self._domain        = domain


    def elaborate(self, platform):
        m = Module()

        phy     = self.phy
        ratio   = self.width // phy.width

        #
        # Clocking and resets.
        #
        m.domains.phy = ClockDomain(local=True, async_reset=True)
        m.d.comb += [
            phy.clk             .eq(self.clk),
            ClockSignal("phy")  .eq(phy.pclk),
        ]

        m.submodules += ResetSynchronizer(phy.reset, domain="phy")
        m.d.comb += [
            phy.reset           .eq(self.reset),
        ]


        #
        # Common gearbox signals.
        #
        if ratio == 1:
            gear_index   = Const(0)
            gear_advance = Const(1)
        else:
            gear_index   = Signal(range(ratio))
            gear_advance = Signal()
            m.d.phy  += gear_index   .eq(gear_index + 1)
            m.d.comb += gear_advance .eq(gear_index == ratio - 1)


        #
        # Transmit data bus and related control signals.
        #
        geared_tx_data          = Signal.like(self.tx_data)
        geared_tx_datak         = Signal.like(self.tx_datak)
        geared_tx_compliance    = Signal.like(self.tx_compliance)
        geared_tx_ones_zeroes   = Signal.like(self.tx_ones_zeroes)
        mac_tx_bus_signals = Cat(
            self.tx_data,
            self.tx_datak,
            # These control signals are additional inputs to the 8b10b encoder; and must be
            # exactly synchronized to the transmit data bus.
            self.tx_compliance,
            self.tx_ones_zeroes,
        )
        phy_tx_bus_signals = Cat(
            geared_tx_data,
            geared_tx_datak,
            geared_tx_compliance,
            geared_tx_ones_zeroes,
        )

        m.d.comb += [
            phy.tx_data         .eq(geared_tx_data .word_select(gear_index, len(phy.tx_data))),
            phy.tx_datak        .eq(geared_tx_datak.word_select(gear_index, len(phy.tx_datak)))
        ]
        # TxCompliance affects only the first transmitted symbol; keep that property after gearing.
        with m.If(gear_index == 0):
            m.d.comb += phy.tx_compliance   .eq(geared_tx_compliance)
        # TxOnesZeroes replaces all symbols on the transmit data bus.
        m.d.comb += phy.tx_ones_zeroes  .eq(geared_tx_ones_zeroes)

        m.submodules.tx_fifo = tx_fifo = AsyncFIFOBuffered(
            width=len(mac_tx_bus_signals),
            depth=4,
            w_domain="sync",
            r_domain="phy"
        )
        m.d.comb += [
            tx_fifo.w_data      .eq(mac_tx_bus_signals),
            tx_fifo.w_en        .eq(1),
            phy_tx_bus_signals  .eq(tx_fifo.r_data),
            tx_fifo.r_en        .eq(gear_advance)
        ]


        #
        # Receive data bus and related control signals.
        #
        geared_rx_data          = [Signal.like(phy.rx_data,     name_suffix=f"_{i}") for i in range(ratio)]
        geared_rx_datak         = [Signal.like(phy.rx_datak,    name_suffix=f"_{i}") for i in range(ratio)]
        geared_rx_valid         = [Signal.like(phy.rx_valid,    name_suffix=f"_{i}") for i in range(ratio)]
        geared_rx_status        = [Signal.like(phy.rx_status,   name_suffix=f"_{i}") for i in range(ratio)]
        geared_phy_status       = [Signal.like(phy.phy_status,  name_suffix=f"_{i}") for i in range(ratio)]
        phy_rx_bus_signals = Cat(
            phy.rx_data,
            phy.rx_datak,
            phy.rx_valid,
            phy.rx_status,
            phy.phy_status,
        )
        mac_rx_bus_signals = Cat((
            geared_rx_data[i],
            geared_rx_datak[i],
            # These control signals are additional outputs from the 8b10b decoder, comma aligner,
            # and elastic buffer; and must be exactly synchronized to the receive data bus.
            geared_rx_valid[i],
            geared_rx_status[i],
            # The PhyStatus signal will be asserted on the same cycle as RxStatus; and must be
            # exactly synchronized with it (and so the data bus, even though it has no direct
            # relationship with the latter).
            geared_phy_status[i],
        ) for i in range(ratio))

        m.d.comb += [
            self.rx_data        .eq(Cat(geared_rx_data)),
            self.rx_datak       .eq(Cat(geared_rx_datak))
        ]
        # RxValid is asserted if all of the symbols are valid.
        with m.If(Cat(geared_rx_valid).any()):
            m.d.comb += self.rx_valid       .eq(1)
        # Several different conditions can be indicated for different symbols transferred over
        # RxData/RxDataK; when this happens, the condition with the highest priority is indicated.
        # The complete priority order (lowest to highest) is:
        #   8. (Received data OK)
        #   7. Receiver detected
        #   6. SKP symbols removed
        #   5. SKP symbols added
        #   4. Disparity errors
        #   3. Elastic buffer underflow
        #   2. Elastic buffer overflow
        #   1. 8B/10B decode error
        for rx_status_code in 0b011, 0b010, 0b001, 0b111, 0b110, 0b101, 0b100:
            with m.If(Cat(geared_rx_status[i] == rx_status_code for i in range(ratio)).any()):
                m.d.comb += self.rx_status      .eq(rx_status_code)
        # PhyStatus is asserted for one cycle once the PHY completes a request submitted by
        # the MAC (and so will never be asserted for two adjacent cycles); or asserted continuously
        # during certain power state transitions (in which case it is unimportant for which exact
        # symbols it is asserted).
        with m.If(Cat(geared_phy_status).any()):
            m.d.comb += self.phy_status     .eq(1)

        m.submodules.rx_fifo = rx_fifo = AsyncFIFOBuffered(
            width=len(mac_rx_bus_signals),
            depth=4,
            w_domain="phy",
            r_domain="sync"
        )
        m.d.phy  += [
            rx_fifo.w_data.word_select(gear_index, len(phy_rx_bus_signals))
                                .eq(phy_rx_bus_signals),
            rx_fifo.w_en        .eq(gear_advance),
        ]
        m.d.comb += [
            mac_rx_bus_signals  .eq(rx_fifo.r_data),
            rx_fifo.r_en        .eq(1)
        ]


        #
        # Control and status signals not related to the data bus.
        #
        m.submodules += [
            # These control and status signals may change while the PHY is active; but they
            # do not need to be precisely synchronized to the data bus, since no specific
            # latency is defined for these signals.
            FFSynchronizer(self.phy_mode,       phy.phy_mode,       o_domain="phy"),
            FFSynchronizer(self.elas_buf_mode,  phy.elas_buf_mode,  o_domain="phy"),
            FFSynchronizer(self.rate,           phy.rate,           o_domain="phy"),
            FFSynchronizer(self.power_down,     phy.power_down,     o_domain="phy"),
            FFSynchronizer(self.tx_deemph,      phy.tx_deemph,      o_domain="phy"),
            FFSynchronizer(self.tx_margin,      phy.tx_margin,      o_domain="phy"),
            FFSynchronizer(self.tx_swing,       phy.tx_swing,       o_domain="phy"),
            FFSynchronizer(self.tx_detrx_lpbk,  phy.tx_detrx_lpbk,  o_domain="phy"),
            FFSynchronizer(self.tx_elec_idle,   phy.tx_elec_idle,   o_domain="phy"),
            FFSynchronizer(self.rx_polarity,    phy.rx_polarity,    o_domain="phy"),
            FFSynchronizer(self.rx_eq_training, phy.rx_eq_training, o_domain="phy"),
            FFSynchronizer(self.rx_termination, phy.rx_termination, o_domain="phy"),
            FFSynchronizer(phy.rx_elec_idle,    self.rx_elec_idle,  o_domain="sync"),
            FFSynchronizer(phy.power_present,   self.power_present, o_domain="sync"),
        ]


        # Rename the default domain to the MAC domain that was requested.
        return DomainRenamer({"sync": self._domain})(m)



class GearedPIPEInterface(Elaboratable):
    """ Module that presents a post-gearing PIPE interface, performing gearing in I/O hardware.

    This module presents a public interface that's identical to a standard PIPE PHY,
    with the following exceptions:

        - ``tx_data``    is 32 bits wide, rather than 16
        - ``tx_datak``   is 4  bits wide, rather than  2
        - ``rx_data``    is 32 bits wide, rather than 16
        - ``rx_datak``   is 4  bits wide, rather than  2
        - ``phy_status`` is 2 bits wide, rather than 1
        - ``rx_status``  is now an array of two 3-bit signals

    This module *requires* that a half-rate / 125MHz clock that's in phase with the ``pipe_io``
    clock be provided to the ``pipe`` domain. This currently must be handled per-device, so it
    is the responsibility of the platform's clock domain generator.

    This module optionally can connect the PIPE I/O clock (``pclk``) to the design's clocking
    network. This configuration is recommended.


    Parameters
    ----------
    pipe: PIPE I/O resource
        The raw PIPE interface to be worked with.
    handle_clocking: boolean, optional
        If true or not provided, this module will attempt to handle some clock connections
        for you. This means that ClockSignal("pipe_io") will be automatically driven by the
        PHY's clock (``pclk``), and ``tx_clk`` will automatically be tied to ClockSignal("pipe").
    """

    # Provide standard XDR settings that can be used when requesting an interface.
    GEARING_XDR = {
        'tx_data': 2, 'tx_datak': 2, 'tx_clk':   2,
        'rx_data': 2, 'rx_datak': 2, 'rx_valid': 2,
        'phy_status': 2, 'rx_status': 2
    }


    def __init__(self, *, pipe, invert_rx_polarity_signal=False):
        self._io = pipe
        self._invert_rx_polarity_signal = invert_rx_polarity_signal

        #
        # I/O port
        #

        self.tx_clk      = Signal()
        self.tx_data     = Signal(32)
        self.tx_datak    = Signal(4)

        self.pclk        = Signal()
        self.rx_data     = Signal(32)
        self.rx_datak    = Signal(4)
        self.rx_valid    = Signal()

        self.phy_status  = Signal(2)
        self.rx_status   = Array((Signal(3), Signal(3)))

        self.rx_polarity = Signal()

        # Copy each of the elements from our core PIPE PHY, so we act mostly as a passthrough.
        for name, *_ in self._io.layout:

            # If we're handling the relevant signal manually, skip it.
            if hasattr(self, name):
                continue

            # Grab the raw I/O...
            io = getattr(self._io, name)

            # If it's a tri-state, copy it as-is.
            if hasattr(io, 'oe'):
                setattr(self, name, io)

            # Otherwise, copy either its input...
            elif hasattr(io,  'i'):
                setattr(self, name, io.i)

            # ... or its output.
            elif hasattr(io,  'o'):
                setattr(self, name, io.o)

            else:
                raise ValueError(f"Unexpected signal {name} with subordinates {io} in PIPE PHY!")


    def elaborate(self, platform):
        m = Module()


        m.d.comb += [
            # Drive our I/O boundary clock with our PHY clock directly,
            # and replace our geared clock with the relevant divided clock.
            ClockSignal("ss_io")     .eq(self._io.pclk.i),
            self.pclk                .eq(ClockSignal("ss")),

            # Drive our transmit clock with an DDR output driven from our full-rate clock.
            # Re-creating the clock in this I/O cell ensures that our clock output is phase-aligned
            # with the signals we create below. [UG471: pg128, "Clock Forwarding"]
            self._io.tx_clk.o_clk    .eq(ClockSignal("ss_io_shifted")),
            self._io.tx_clk.o0       .eq(1),
            self._io.tx_clk.o1       .eq(0),
        ]

        # Set up our geared I/O clocks.
        m.d.comb += [
            # Drive our transmit signals from our transmit-domain clocks...
            self._io.tx_data.o_clk     .eq(ClockSignal("ss")),
            self._io.tx_datak.o_clk    .eq(ClockSignal("ss")),

            # ... and drive our receive signals from our primary/receive domain clock.
            self._io.rx_data.i_clk     .eq(ClockSignal("ss_shifted")),
            self._io.rx_datak.i_clk    .eq(ClockSignal("ss_shifted")),
            self._io.rx_valid.i_clk    .eq(ClockSignal("ss_shifted")),
            self._io.phy_status.i_clk  .eq(ClockSignal("ss_shifted")),
            self._io.rx_status.i_clk   .eq(ClockSignal("ss_shifted")),
        ]

        #
        # Output handling.
        #
        m.d.ss += [
            # We'll output tx_data bytes {0, 1} and _then_ {2, 3}.
            self._io.tx_data.o0   .eq(self.tx_data [ 0:16]),
            self._io.tx_data.o1   .eq(self.tx_data [16:32]),
            self._io.tx_datak.o0  .eq(self.tx_datak[ 0: 2]),
            self._io.tx_datak.o1  .eq(self.tx_datak[ 2: 4])
        ]

        #
        # Input handling.
        #
        m.d.ss += [
            # We'll capture rx_data bytes {0, 1} and _then_ {2, 3}.
            self.rx_data [ 0:16]  .eq(self._io.rx_data.i0),
            self.rx_data [16:32]  .eq(self._io.rx_data.i1),
            self.rx_datak[ 0: 2]  .eq(self._io.rx_datak.i0),
            self.rx_datak[ 2: 4]  .eq(self._io.rx_datak.i1),

            # Split our RX_STATUS to march our other geared I/O.
            self.phy_status[0]    .eq(self._io.phy_status.i0),
            self.phy_status[1]    .eq(self._io.phy_status.i1),

            self.rx_status[0]     .eq(self._io.rx_status.i0),
            self.rx_status[1]     .eq(self._io.rx_status.i1),


            # RX_VALID indicates that we have symbol lock; and thus should remain
            # high throughout our whole stream. Accordingly, we can squish both values
            # down into a single value without losing anything, as it should remain high
            # once our signal has been trained.
            self.rx_valid         .eq(self._io.rx_valid.i0 & self._io.rx_valid.i1),
        ]



        # Allow us to invert the polarity of our ``rx_polarity`` signal, to account for
        # boards that have their Rx+/- lines swapped.
        if self._invert_rx_polarity_signal:
            m.d.comb += self._io.rx_polarity.eq(~self.rx_polarity)
        else:
            m.d.comb += self._io.rx_polarity.eq(self.rx_polarity)

        return m





#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Request components shared between USB2 and USB3. """

from amaranth       import *
from amaranth.hdl.rec import DIR_FANOUT


class SetupPacket(Record):
    """ Record capturing the content of a setup packet.

    Components (O = output from setup parser; read-only input to others):
        O: received      -- Strobe; indicates that a new setup packet has been received,
                            and thus this data has been updated.

        O: is_in_request -- High if the current request is an 'in' request.
        O: type[2]       -- Request type for the current request.
        O: recipient[5]  -- Recipient of the relevant request.

        O: request[8]    -- Request number.
        O: value[16]     -- Value argument for the setup request.
        O: index[16]     -- Index argument for the setup request.
        O: length[16]    -- Length of the relevant setup request.
    """

    def __init__(self):
        super().__init__([
            # Byte 1
            ('recipient',      5, DIR_FANOUT),
            ('type',           2, DIR_FANOUT),
            ('is_in_request',  1, DIR_FANOUT),

            # Byte 2
            ('request',        8, DIR_FANOUT),

            # Byte 3/4
            ('value',         16, DIR_FANOUT),

            # Byte 5/6
            ('index',         16, DIR_FANOUT),

            # Byte 7/8
            ('length',        16, DIR_FANOUT),

            # Control signaling.
            ('received',       1, DIR_FANOUT),
        ])




# BSD 3-Clause License
#
# Adapted from ValentyUSB.
#
# Copyright (c) 2020, Great Scott Gadgets <ktemkin@greatscottgadgets.com>
# Copyright (c) 2018, Luke Valenty
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from amaranth          import Elaboratable, Module, Signal, Cat, Const, ClockSignal
from amaranth.lib.cdc  import FFSynchronizer
from amaranth.lib.fifo import AsyncFIFOBuffered
from amaranth.hdl.xfrm import ResetInserter

class RxClockDataRecovery(Elaboratable):
    """RX Clock Data Recovery module.

    RxClockDataRecovery synchronizes the USB differential pair with the FPGAs
    clocks, de-glitches the differential pair, and recovers the incoming clock
    and data.

    Clock Domain
    ------------
    usb_48 : 48MHz

    Input Ports
    -----------
    Input ports are passed in via the constructor.

    usbp_raw : Signal(1)
        Raw USB+ input from the FPGA IOs, no need to synchronize.

    usbn_raw : Signal(1)
        Raw USB- input from the FPGA IOs, no need to synchronize.

    Output Ports
    ------------
    Output ports are data members of the module. All output ports are flopped.
    The line_state_dj/dk/se0/se1 outputs are 1-hot encoded.

    line_state_valid : Signal(1)
        Asserted for one clock when the output line state is ready to be sampled.

    line_state_dj : Signal(1)
        Represents Full Speed J-state on the incoming USB data pair.
        Qualify with line_state_valid.

    line_state_dk : Signal(1)
        Represents Full Speed K-state on the incoming USB data pair.
        Qualify with line_state_valid.

    line_state_se0 : Signal(1)
        Represents SE0 on the incoming USB data pair.
        Qualify with line_state_valid.

    line_state_se1 : Signal(1)
        Represents SE1 on the incoming USB data pair.
        Qualify with line_state_valid.
    """
    def __init__(self, usbp_raw, usbn_raw):
        self._usbp = usbp_raw
        self._usbn = usbn_raw


        self.line_state_valid = Signal()
        self.line_state_dj = Signal()
        self.line_state_dk = Signal()
        self.line_state_se0 = Signal()
        self.line_state_se1 = Signal()



    def elaborate(self, platform):
        m = Module()

        # Synchronize the USB signals at our I/O boundary.
        # Despite the assumptions made in ValentyUSB, this line rate recovery FSM
        # isn't enough to properly synchronize these inputs. We'll explicitly synchronize.
        sync_dp = Signal()
        sync_dn = Signal()
        m.submodules.dp_cdc = FFSynchronizer(self._usbp, sync_dp, o_domain="usb_io")
        m.submodules.dn_cdc = FFSynchronizer(self._usbn, sync_dn, o_domain="usb_io")

        #######################################################################
        # Line State Recovery State Machine
        #
        # The receive path doesn't use a differential receiver.  Because of
        # this there is a chance that one of the differential pairs will appear
        # to have changed to the new state while the other is still in the old
        # state.  The following state machine detects transitions and waits an
        # extra sampling clock before decoding the state on the differential
        # pair.  This transition period  will only ever last for one clock as
        # long as there is no noise on the line.  If there is enough noise on
        # the line then the data may be corrupted and the packet will fail the
        # data integrity checks.
        #
        dpair =  Cat(sync_dp, sync_dn)

        # output signals for use by the clock recovery stage
        line_state_in_transition = Signal()

        with m.FSM(domain="usb_io") as fsm:
            m.d.usb_io += [
                self.line_state_se0  .eq(fsm.ongoing("SE0")),
                self.line_state_se1  .eq(fsm.ongoing("SE1")),
                self.line_state_dj   .eq(fsm.ongoing("DJ" )),
                self.line_state_dk   .eq(fsm.ongoing("DK" )),
            ]

            # If we are in a transition state, then we can sample the pair and
            # move to the next corresponding line state.
            with m.State("DT"):
                m.d.comb += line_state_in_transition.eq(1)

                with m.Switch(dpair):
                    with m.Case(0b10):
                        m.next = "DJ"
                    with m.Case(0b01):
                        m.next = "DK"
                    with m.Case(0b00):
                        m.next = "SE0"
                    with m.Case(0b11):
                        m.next = "SE1"

            # If we are in a valid line state and the value of the pair changes,
            # then we need to move to the transition state.
            with m.State("DJ"):
                with m.If(dpair != 0b10):
                    m.next = "DT"

            with m.State("DK"):
                with m.If(dpair != 0b01):
                    m.next = "DT"

            with m.State("SE0"):
                with m.If(dpair != 0b00):
                    m.next = "DT"

            with m.State("SE1"):
                with m.If(dpair != 0b11):
                    m.next = "DT"


        #######################################################################
        # Clock and Data Recovery
        #
        # The DT state from the line state recovery state machine is used to align to
        # transmit clock.  The line state is sampled in the middle of the bit time.
        #
        # Example of signal relationships
        # -------------------------------
        # line_state        DT  DJ  DJ  DJ  DT  DK  DK  DK  DK  DK  DK  DT  DJ  DJ  DJ
        # line_state_valid  ________----____________----____________----________----____
        # bit_phase         0   0   1   2   3   0   1   2   3   0   1   2   0   1   2
        #

        # We 4x oversample, so make the line_state_phase have
        # 4 possible values.
        line_state_phase = Signal(2)
        m.d.usb_io += self.line_state_valid.eq(line_state_phase == 1)


        with m.If(line_state_in_transition):
            m.d.usb_io += [
                # re-align the phase with the incoming transition
                line_state_phase.eq(0),

                # make sure we never assert valid on a transition
                self.line_state_valid.eq(0),
            ]
        with m.Else():
            # keep tracking the clock by incrementing the phase
            m.d.usb_io += line_state_phase.eq(line_state_phase + 1)

        return m


class RxNRZIDecoder(Elaboratable):
    """RX NRZI decoder.

    In order to ensure there are enough bit transitions for a receiver to recover
    the clock usb uses NRZI encoding.  This module processes the incoming
    dj, dk, se0, and valid signals and decodes them to data values.  It
    also pipelines the se0 signal and passes it through unmodified.

    https://www.pjrc.com/teensy/beta/usb20.pdf, USB2 Spec, 7.1.8
    https://en.wikipedia.org/wiki/Non-return-to-zero

    Clock Domain
    ------------
    usb_48 : 48MHz

    Input Ports
    -----------
    Input ports are passed in via the constructor.

    i_valid : Signal(1)
        Qualifier for all of the input signals.  Indicates one bit of valid
        data is present on the inputs.

    i_dj : Signal(1)
        Indicates the bus is currently in a Full-Speed J-state.
        Qualified by valid.

    i_dk : Signal(1)
        Indicates the bus is currently in a Full-Speed K-state.
        Qualified by valid.

    i_se0 : Signal(1)
        Indicates the bus is currently in a SE0 state.
        Qualified by valid.

    Output Ports
    ------------
    Output ports are data members of the module. All output ports are flopped.

    o_valid : Signal(1)
        Qualifier for all of the output signals. Indicates one bit of valid
        data is present on the outputs.

    o_data : Signal(1)
        Decoded data bit from USB bus.
        Qualified by valid.

    o_se0 : Signal(1)
        Indicates the bus is currently in a SE0 state.
        Qualified by valid.
    """

    def __init__(self):
        self.i_valid = Signal()
        self.i_dj = Signal()
        self.i_dk = Signal()
        self.i_se0 = Signal()

        # pass all of the outputs through a pipe stage
        self.o_valid = Signal(1)
        self.o_data = Signal(1)
        self.o_se0 = Signal(1)


    def elaborate(self, platform):
        m = Module()

        last_data = Signal()
        with m.If(self.i_valid):
            m.d.usb_io += [
                last_data.eq(self.i_dk),
                self.o_data.eq(~(self.i_dk ^ last_data)),
                self.o_se0.eq((~self.i_dj) & (~self.i_dk)),
            ]
        m.d.usb_io += self.o_valid.eq(self.i_valid),

        return m


class RxPacketDetect(Elaboratable):
    """Packet Detection

    Full Speed packets begin with the following sequence:

        KJKJKJKK

    This raw sequence corresponds to the following data:

        00000001

    The bus idle condition is signaled with the J state:

        JJJJJJJJ

    This translates to a series of '1's since there are no transitions.  Given
    this information, it is easy to detect the beginning of a packet by looking
    for 00000001.

    The end of a packet is even easier to detect.  The end of a packet is
    signaled with two SE0 and one J.  We can just look for the first SE0 to
    detect the end of the packet.

    Packet detection can occur in parallel with bitstuff removal.

    https://www.pjrc.com/teensy/beta/usb20.pdf, USB2 Spec, 7.1.10

    Input Ports
    ------------
    i_valid : Signal(1)
        Qualifier for all of the input signals.  Indicates one bit of valid
        data is present on the inputs.

    i_data : Signal(1)
        Decoded data bit from USB bus.
        Qualified by valid.

    i_se0 : Signal(1)
        Indicator for SE0 from USB bus.
        Qualified by valid.

    Output Ports
    ------------
    o_pkt_start : Signal(1)
        Asserted for one clock on the last bit of the sync.

    o_pkt_active : Signal(1)
        Asserted while in the middle of a packet.

    o_pkt_end : Signal(1)
        Asserted for one clock after the last data bit of a packet was received.
    """

    def __init__(self):
        self.i_valid = Signal()
        self.i_data = Signal()
        self.i_se0 = Signal()

        self.o_pkt_start = Signal()
        self.o_pkt_active = Signal()
        self.o_pkt_end = Signal()


    def elaborate(self, platform):
        m = Module()

        pkt_start = Signal()
        pkt_active = Signal()
        pkt_end = Signal()

        with m.FSM(domain="usb_io"):

            for i in range(5):

                with m.State(f"D{i}"):
                    with m.If(self.i_valid):
                        with m.If(self.i_data | self.i_se0):
                            # Receiving '1' or SE0 early resets the packet start counter.
                            m.next = "D0"

                        with m.Else():
                            # Receiving '0' increments the packet start counter.
                            m.next = f"D{i + 1}"

            with m.State("D5"):
                with m.If(self.i_valid):
                    with m.If(self.i_se0):
                        m.next = "D0"
                    # once we get a '1', the packet is active
                    with m.Elif(self.i_data):
                        m.d.comb += pkt_start.eq(1)
                        m.next = "PKT_ACTIVE"

            with m.State("PKT_ACTIVE"):
                m.d.comb += pkt_active.eq(1)
                with m.If(self.i_valid & self.i_se0):
                    m.d.comb += [
                        pkt_active.eq(0),
                        pkt_end.eq(1)
                    ]
                    m.next = "D0"

        # pass all of the outputs through a pipe stage
        m.d.comb += [
            self.o_pkt_start.eq(pkt_start),
            self.o_pkt_active.eq(pkt_active),
            self.o_pkt_end.eq(pkt_end),
        ]

        return m



class RxBitstuffRemover(Elaboratable):
    """RX Bitstuff Removal

    Long sequences of 1's would cause the receiver to lose it's lock on the
    transmitter's clock.  USB solves this with bitstuffing.  A '0' is stuffed
    after every 6 consecutive 1's.  This extra bit is required to recover the
    clock, but it should not be passed on to higher layers in the device.

    https://www.pjrc.com/teensy/beta/usb20.pdf, USB2 Spec, 7.1.9
    https://en.wikipedia.org/wiki/Bit_stuffing

    Clock Domain
    ------------
    usb_48 : 48MHz

    Input Ports
    ------------
    i_valid : Signal(1)
        Qualifier for all of the input signals.  Indicates one bit of valid
        data is present on the inputs.

    i_data : Signal(1)
        Decoded data bit from USB bus.
        Qualified by valid.

    Output Ports
    ------------
    o_data : Signal(1)
        Decoded data bit from USB bus.

    o_stall : Signal(1)
        Indicates the bit stuffer just removed an extra bit, so no data available.

    o_error : Signal(1)
        Indicates there has been a bitstuff error. A bitstuff error occurs
        when there should be a stuffed '0' after 6 consecutive 1's; but instead
        of a '0', there is an additional '1'.  This is normal during IDLE, but
        should never happen within a packet.
        Qualified by valid.
    """

    def __init__(self):
        self.i_valid = Signal()
        self.i_data = Signal()

        # pass all of the outputs through a pipe stage
        self.o_data = Signal()
        self.o_error = Signal()
        self.o_stall = Signal(init=1)


    def elaborate(self, platform):
        m = Module()

        # This state machine recognizes sequences of 6 bits and drops the 7th
        # bit.  The fsm implements a counter in a series of several states.
        # This is intentional to help absolutely minimize the levels of logic
        # used.
        drop_bit = Signal(1)


        with m.FSM(domain="usb_io"):

            for i in range(6):
                with m.State(f"D{i}"):
                    with m.If(self.i_valid):
                        with m.If(self.i_data):
                            # Receiving '1' increments the bitstuff counter.
                            m.next = (f"D{i + 1}")
                        with m.Else():
                            # Receiving '0' resets the bitstuff counter.
                            m.next = "D0"

            with m.State("D6"):
                with m.If(self.i_valid):
                    m.d.comb += drop_bit.eq(1)
                    # Reset the bitstuff counter, drop the data.
                    m.next = "D0"

        m.d.usb_io += [
            self.o_data.eq(self.i_data),
            self.o_stall.eq(drop_bit | ~self.i_valid),
            self.o_error.eq(drop_bit & self.i_data & self.i_valid),
        ]

        return m


class RxShifter(Elaboratable):
    """RX Shifter

    A shifter is responsible for shifting in serial bits and presenting them
    as parallel data.  The shifter knows how many bits to shift and has
    controls for resetting the shifter.

    Clock Domain
    ------------
    usb    : 12MHz

    Parameters
    ----------
    Parameters are passed in via the constructor.

    width : int
        Number of bits to shift in.

    Input Ports
    -----------
    i_valid : Signal(1)
        Qualifier for all of the input signals.  Indicates one bit of valid
        data is present on the inputs.

    i_data : Signal(1)
        Serial input data.
        Qualified by valid.

    Output Ports
    ------------
    o_data : Signal(width)
        Shifted in data.

    o_put : Signal(1)
        Asserted for one clock once the register is full.
    """
    def __init__(self, width):
        self._width = width

        #
        # I/O port
        #
        self.reset   = Signal()

        self.i_valid = Signal()
        self.i_data  = Signal()

        self.o_data  = Signal(width)
        self.o_put   = Signal()


    def elaborate(self, platform):
        m = Module()
        width = self._width

        # Instead of using a counter, we will use a sentinel bit in the shift
        # register to indicate when it is full.
        shift_reg = Signal(width+1, init=0b1)

        m.d.comb += self.o_data.eq(shift_reg[0:width])
        m.d.usb_io += self.o_put.eq(shift_reg[width-1] & ~shift_reg[width] & self.i_valid),

        with m.If(self.reset):
            m.d.usb_io += shift_reg.eq(1)

        with m.If(self.i_valid):
            with m.If(shift_reg[width]):
                m.d.usb_io += shift_reg.eq(Cat(self.i_data, Const(1)))
            with m.Else():
                m.d.usb_io += shift_reg.eq(Cat(self.i_data, shift_reg[0:width])),

        return m


class RxPipeline(Elaboratable):

    def __init__(self):
        self.reset = Signal()

        # 12MHz USB alignment pulse in 48MHz clock domain
        self.o_bit_strobe = Signal()

        # Reset state is J
        self.i_usbp = Signal(init=1)
        self.i_usbn = Signal(init=0)

        self.o_data_strobe = Signal()
        self.o_data_payload = Signal(8)

        self.o_pkt_start = Signal()
        self.o_pkt_in_progress = Signal()
        self.o_pkt_end = Signal()

        self.o_receive_error = Signal()


    def elaborate(self, platform):
        m = Module()

        #
        # Clock/Data recovery.
        #
        clock_data_recovery = RxClockDataRecovery(self.i_usbp, self.i_usbn)
        m.submodules.clock_data_recovery = clock_data_recovery
        m.d.comb += self.o_bit_strobe.eq(clock_data_recovery.line_state_valid)

        #
        # NRZI decoding
        #
        m.submodules.nrzi = nrzi = RxNRZIDecoder()
        m.d.comb += [
            nrzi.i_valid  .eq(self.o_bit_strobe),
            nrzi.i_dj     .eq(clock_data_recovery.line_state_dj),
            nrzi.i_dk     .eq(clock_data_recovery.line_state_dk),
            nrzi.i_se0    .eq(clock_data_recovery.line_state_se0),
        ]

        #
        # Packet boundary detection.
        #
        m.submodules.detect = detect = ResetInserter(self.reset)(RxPacketDetect())
        m.d.comb += [
            detect.i_valid  .eq(nrzi.o_valid),
            detect.i_se0    .eq(nrzi.o_se0),
            detect.i_data   .eq(nrzi.o_data),
        ]

        #
        # Bitstuff remover.
        #
        m.submodules.bitstuff = bitstuff = \
            ResetInserter(~detect.o_pkt_active)(RxBitstuffRemover())
        m.d.comb += [
            bitstuff.i_valid.eq(nrzi.o_valid),
            bitstuff.i_data.eq(nrzi.o_data),
            self.o_receive_error.eq(bitstuff.o_error)
        ]

        #
        # 1bit->8bit (1byte) gearing
        #
        m.submodules.shifter = shifter = RxShifter(width=8)
        past_o_pkt_active    = Signal()
        m.d.usb_io          += past_o_pkt_active.eq(detect.o_pkt_active)
        m.d.comb += [
            shifter.reset.eq(detect.o_pkt_end),
            shifter.i_data.eq(bitstuff.o_data),
            shifter.i_valid.eq(~bitstuff.o_stall & past_o_pkt_active),
        ]

        #
        # Clock domain crossing.
        #
        flag_start  = Signal()
        flag_end    = Signal()
        flag_valid  = Signal()
        m.submodules.payload_fifo = payload_fifo = AsyncFIFOBuffered(width=8, depth=4, r_domain="usb", w_domain="usb_io")
        m.d.comb += [
            payload_fifo.w_data  .eq(shifter.o_data[::-1]),
            payload_fifo.w_en    .eq(shifter.o_put),
            self.o_data_payload  .eq(payload_fifo.r_data),
            self.o_data_strobe   .eq(payload_fifo.r_rdy),
            payload_fifo.r_en    .eq(1)
        ]

        m.submodules.flags_fifo = flags_fifo = AsyncFIFOBuffered(width=2, depth=4, r_domain="usb", w_domain="usb_io")
        m.d.comb += [
            flags_fifo.w_data[1]  .eq(detect.o_pkt_start),
            flags_fifo.w_data[0]  .eq(detect.o_pkt_end),
            flags_fifo.w_en       .eq(detect.o_pkt_start | detect.o_pkt_end),

            flag_start            .eq(flags_fifo.r_data[1]),
            flag_end              .eq(flags_fifo.r_data[0]),
            flag_valid            .eq(flags_fifo.r_rdy),
            flags_fifo.r_en       .eq(1),
        ]

        # Packet flag signals (in 12MHz domain)
        m.d.comb += [
            self.o_pkt_start  .eq(flag_start & flag_valid),
            self.o_pkt_end    .eq(flag_end & flag_valid),
        ]

        with m.If(self.o_pkt_start):
            m.d.usb += self.o_pkt_in_progress.eq(1)
        with m.Elif(self.o_pkt_end):
            m.d.usb += self.o_pkt_in_progress.eq(0)

        return m




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Link Management Packet (LMP) -related gateware. """

from amaranth import *
from usb_protocol.types.superspeed import HeaderPacketType, LinkManagementPacketSubtype

from ..link.header import HeaderQueue, HeaderPacket


class PortCapabilityHeaderPacket(HeaderPacket):
    DW0_LAYOUT = [
        ('type',        5),
        ('subtype',     4),
        ('link_speed',  7),
        ('reserved',   16)
    ]
    DW1_LAYOUT = [
        ('num_hp_buffers',      8),
        ('reserved_0',          8),
        ('supports_downstream', 1),
        ('supports_upstream',   1),
        ('reserved_1',          2),
        ('tiebreaker',          4),
        ('reserved_2',          8),
    ]


class PortConfigurationHeaderPacket(HeaderPacket):
    DW0_LAYOUT = [
        ('type',        5),
        ('subtype',     4),
        ('link_speed',  7),
        ('reserved',   16)
    ]


class PortConfigurationResponseHeaderPacket(HeaderPacket):
    DW0_LAYOUT = [
        ('type',           5),
        ('subtype',        4),
        ('response_code',  7),
        ('reserved',      16)
    ]




class LinkManagementPacketHandler(Elaboratable):
    """ Gateware that handles Link Management Packets.

    Attributes
    -----------
    header_sink: HeaderQueue(), input stream
        Stream that brings up header packets for handling.
    header_source: HeaderQueue(), output stream
        Stream that accepts header packets for generation.

    link_ready: Signal(), input
        Should be asserted once our link is ready; used to trigger advertising.
    """

    # Link speed constants.
    LINK_SPEED_5GBPS = 1

    # Constants.
    CONFIGURATION_ACCEPTED = 1
    CONFIGURATION_REJECTED = 2

    def __init__(self):

        #
        # I/O port
        #
        self.header_sink   = HeaderQueue()
        self.header_source = HeaderQueue()

        # Status / control.
        self.usb_reset     = Signal()
        self.link_ready    = Signal()


    def elaborate(self, platform):
        m = Module()

        header_sink   = self.header_sink
        header_source = self.header_source


        #
        # Pending "tasks" for our transmitter.
        #
        pending_configuration_result = Signal(2)


        #
        # LMP transmitter.
        #

        def send_packet_response(response_type, **fields):
            """ Helper that allows us to easily define a packet-send state."""

            # Create a response packet, and mark ourselves as sending it.
            response = response_type()
            m.d.comb += [
                header_source.valid    .eq(1),
                header_source.header   .eq(response),

                response.type          .eq(HeaderPacketType.LINK_MANAGEMENT)
            ]

            # Next, fill in each of the fields:
            for field, value in fields.items():
                m.d.comb += response[field].eq(value)


        def handle_resets():
            """ Helper that brings down the link on USB reset. """
            with m.If(self.usb_reset):
                m.next = "LINK_DOWN"


        with m.FSM(domain="ss"):

            # LINK_DOWN -- our link is not yet ready to exchange packets; we'll wait until
            # it's come up to the point where we can exchange header packets.
            with m.State("LINK_DOWN"):

                # Once our link is ready, we're ready to start link bringup.
                with m.If(self.link_ready):
                    m.next = "SEND_CAPABILITIES"
                with m.Else():
                    m.d.ss += pending_configuration_result.eq(0)


            # SEND_CAPABILITIES -- our link has come up; and we're now ready to advertise our link
            # capabilities to the other side of our link [USB3.2r1: 8.4.5].
            with m.State("SEND_CAPABILITIES"):
                handle_resets()

                send_packet_response(PortCapabilityHeaderPacket,
                    subtype           = LinkManagementPacketSubtype.PORT_CAPABILITY,
                    link_speed        = self.LINK_SPEED_5GBPS,

                    # We're required by specification to support exactly four buffers.
                    num_hp_buffers    = 4,

                    # For now, we only can operate as an upstream device.
                    supports_upstream = 1
                )

                # Continue to drive our packet until it's accepted by the link layer.
                with m.If(header_source.ready):
                    m.next = "DISPATCH_COMMANDS"


            # DISPATCH_COMMANDS -- we'll wait for a command to be queued, and then send it.
            with m.State("DISPATCH_COMMANDS"):
                handle_resets()

                # If we have a pending configuration result, send it!
                with m.If(pending_configuration_result):
                    m.next = "SEND_PORT_CONFIGURATION_RESPONSE"


            # SEND_CONFIGURATION_RESPONSE -- we're sending a Port Configuration Response,
            # typically as a result of receiving a Port Configuration Request packet.
            with m.State("SEND_PORT_CONFIGURATION_RESPONSE"):
                handle_resets()

                send_packet_response(PortConfigurationResponseHeaderPacket,
                    subtype       = LinkManagementPacketSubtype.PORT_CONFIGURATION_RESPONSE,
                    response_code = pending_configuration_result
                )

                # Continue to drive our packet until it's accepted by the link layer.
                with m.If(header_source.ready):
                    m.d.ss += pending_configuration_result.eq(0)
                    m.next = "DISPATCH_COMMANDS"


        #
        # LMP receiver.
        #

        # We'll handle all link management packets.
        new_packet = header_sink.valid
        is_for_us  = header_sink.get_type() == HeaderPacketType.LINK_MANAGEMENT

        with m.If(new_packet & is_for_us):

            # Accept the packet from the physical layer, so its buffer will be freed
            # on the next clock cycle.
            m.d.comb += header_sink.ready.eq(1)

            # We'll handle link management packets based on their subtype.
            subtype = header_sink.header.dw0[5:9]
            with m.Switch(subtype):

                # As an upstream-only Gen1 port, there's not much we need to do with
                # capability advertisements. For now, we'll mostly ignore them.
                with m.Case(LinkManagementPacketSubtype.PORT_CAPABILITY):
                    pass

                # If we receive a PORT_CONFIGURATION request, then our host is assigning
                # us a configuration.
                with m.Case(LinkManagementPacketSubtype.PORT_CONFIGURATION):
                    configuration = PortConfigurationHeaderPacket()
                    m.d.comb += configuration.eq(header_sink.header)

                    # For now, we only support Gen1 / 5Gbps, so we'll accept only links
                    # with that speed selected.
                    with m.If(configuration.link_speed == self.LINK_SPEED_5GBPS):
                        m.d.ss += pending_configuration_result.eq(self.CONFIGURATION_ACCEPTED)
                    with m.Else():
                        m.d.ss += pending_configuration_result.eq(self.CONFIGURATION_REJECTED)


                # TODO: handle any invalid packet types?
                with m.Default():
                    pass

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" SPI and derived interfaces. """

from amaranth import Signal, Module, Cat, Elaboratable, Record
from amaranth.hdl.rec import DIR_FANIN, DIR_FANOUT

from ..test.utils import LunaGatewareTestCase, sync_test_case

class SPIBus(Record):
    """ Record representing an SPI bus. """

    def __init__(self):
        super().__init__([
            ('sck', 1, DIR_FANIN),
            ('sdi', 1, DIR_FANIN),
            ('sdo', 1, DIR_FANOUT),
            ('cs',  1, DIR_FANIN)
        ])


class SPIDeviceInterface(Elaboratable):
    """ Simple word-oriented SPI interface.

    I/O signals:
        B: spi           -- the SPI bus to work with

        O: word_in       -- the most recent word received
        O: word_complete -- strobe indicating a new word is present on word_in
        I: word_out      -- the word to be loaded; latched in on next word_complete and while cs is low
    """

    def __init__(self, *, word_size=8, clock_polarity=0, clock_phase=0, msb_first=True, cs_idles_high=False):
        """
        Parameters:
            word_size      -- The size of each transmitted word, in bits.
            clock_polarity -- The SPI-standard clock polarity. 0 for idle low, 1 for idle high.
            clock_phase    -- The SPI-standard clock phase. 1 to capture on the leading edge, or 0 for on the trailing
            msb_first      -- If true, or not provided, data will be transmitted MSB first (standard).
            cs_idles_high  -- If provided, data will be captured when CS goes _low_, rather than high.
        """

        self.word_size      = word_size
        self.clock_polarity = clock_polarity
        self.clock_phase    = clock_phase
        self.msb_first      = msb_first
        self.cs_idles_high  = cs_idles_high

        #
        # I/O port.
        #

        # SPI
        self.spi = SPIBus()

        # Data I/O
        self.word_in        = Signal(self.word_size)
        self.word_out       = Signal(self.word_size)
        self.word_accepted  = Signal()
        self.word_complete  = Signal()


    def spi_edge_detectors(self, m):
        """ Generates edge detectors for the sample and output clocks, based on the current SPI mode.

        Returns:
            sample_edge, output_edge -- signals that pulse high for a single cycle when we should
                                        sample and change our outputs, respectively
        """

        # Select whether we're working with an inverted or un-inverted serial clock.
        serial_clock = Signal()
        if self.clock_polarity:
            m.d.comb += serial_clock.eq(~self.spi.sck)
        else:
            m.d.comb += serial_clock.eq(self.spi.sck)

        # Generate the leading and trailing edge detectors.
        # Note that we use rising and falling edge detectors, but call these leading and
        # trailing edges, as our clock here may have been inverted.
        past_clk      = Signal()
        m.d.sync     += past_clk.eq(serial_clock)
        leading_edge  = ~past_clk &  serial_clock
        trailing_edge =  past_clk & ~serial_clock

        # Determine the sample and output edges based on the SPI clock phase.
        sample_edge = trailing_edge if self.clock_phase else leading_edge
        output_edge = leading_edge if self.clock_phase else trailing_edge

        return sample_edge, output_edge


    def elaborate(self, platform):
        m = Module()

        # Grab signals that detect when we should shift in and out.
        sample_edge, output_edge = self.spi_edge_detectors(m)

        # We'll use separate buffers for transmit and receive,
        # as this makes the code a little more readable.
        bit_count    = Signal(range(0, self.word_size), init=0)
        current_tx   = Signal.like(self.word_out)
        current_rx   = Signal.like(self.word_in)

        # Signal that tracks if this is our first bit of the word.
        is_first_bit = Signal()

        # A word is ready one cycle after we complete a transaction
        # (and latch in the next word to be sent).
        with m.If(self.word_accepted):
            m.d.sync += [
                self.word_in       .eq(current_rx),
                self.word_complete .eq(1)
            ]
        with m.Else():
            m.d.sync += self.word_complete.eq(0)

        # De-assert our control signals unless explicitly asserted.
        m.d.sync += [
            self.word_accepted .eq(0),
        ]

        # If the chip is selected, process our I/O:
        chip_selected = self.spi.cs if not self.cs_idles_high else ~self.spi.cs

        with m.If(chip_selected):

            # Shift in data on each sample edge.
            with m.If(sample_edge):
                m.d.sync += [
                    bit_count    .eq(bit_count + 1),
                    is_first_bit .eq(0)
                ]

                if self.msb_first:
                    m.d.sync += current_rx.eq(Cat(self.spi.sdi, current_rx[:-1]))
                else:
                    m.d.sync += current_rx.eq(Cat(current_rx[1:], self.spi.sdi))

                # If we're just completing a word, handle I/O.
                with m.If(bit_count + 1 == self.word_size):
                    m.d.sync += [
                        self.word_accepted .eq(1),
                        current_tx         .eq(self.word_out)
                    ]


            # Shift out data on each output edge.
            with m.If(output_edge):
                if self.msb_first:
                    m.d.sync += Cat(current_tx[1:], self.spi.sdo).eq(current_tx)
                else:
                    m.d.sync += Cat(self.spi.sdo, current_tx[:-1]).eq(current_tx)

        with m.Else():
            m.d.sync += [
                current_tx   .eq(self.word_out),
                bit_count    .eq(0),
                is_first_bit .eq(1)
            ]

        return m



class SPIGatewareTestCase(LunaGatewareTestCase):
    """ Extended version of the LunaGatewareTestCase.

    Adds three SPI-simulation methods:
        -spi_send_bit
        -spi_exchange_byte
        -spi_exchange_data
    """

    def spi_send_bit(self, bit):
        """ Sends a single bit over the SPI bus. """
        cycles_per_bit = 4
        spi = self.dut.spi

        # Apply the new bit...
        if hasattr(spi, 'sdi'):
            yield spi.sdi.eq(bit)
            yield from self.advance_cycles(cycles_per_bit)

        # Create a RE of our serial clock.
        yield spi.sck.eq(1)
        yield from self.advance_cycles(cycles_per_bit)

        # Read the data on the bus, and then create our falling edge.
        return_value = (yield spi.sdo)
        yield from self.advance_cycles(cycles_per_bit)

        yield spi.sck.eq(0)
        yield from self.advance_cycles(cycles_per_bit)

        return return_value


    def spi_exchange_byte(self, datum, *, msb_first=True):
        """ Sends a by over the virtual SPI bus. """

        bits = "{:08b}".format(datum)
        data_received = ""

        if not msb_first:
            bits = bits[::-1]

        # Send each of our bits...
        for bit in bits:
            received = yield from self.spi_send_bit(int(bit))
            data_received += '1' if received else '0'

        if not msb_first:
            data_received = data_received[::-1]

        return int(data_received, 2)


    def spi_exchange_data(self, data, msb_first=True):
        """ Sends a string of bytes over our virtual SPI bus. """

        yield self.dut.spi.cs.eq(1)
        yield

        response = bytearray()

        for byte in data:
            response_byte = yield from self.spi_exchange_byte(byte)
            response.append(response_byte)

        yield self.dut.spi.cs.eq(0)
        yield

        return response



class SPICommandInterface(Elaboratable):
    """ Variant of an SPIDeviceInterface that accepts command-prefixed data.

    I/O signals:
        I: sck           -- SPI clock, from the SPI master
        I: sdi           -- SPI data in
        O: sdo           -- SPI data out
        I: cs            -- chip select, active high (as we assume your I/O will use PinsN)

        O: command       -- the command read from the SPI bus
        O: command_ready -- a new command is ready

        O: word_received -- the most recent word received
        O: word_complete -- strobe indicating a new word is present on word_in
        I: word_to_send  -- the word to be loaded; latched in on next word_complete and while cs is low

        O: idle          -- true iff the register interface is currently doing nothing
        O: stalled       -- true iff the register interface cannot accept data until this transaction ends
    """

    def __init__(self, command_size=8, word_size=32):

        self.command_size = command_size
        self.word_size    = word_size

        #
        # I/O port.
        #

        # SPI
        self.spi = SPIBus()

        # Command I/O.
        self.command        = Signal(self.command_size)
        self.command_ready  = Signal()

        # Data I/O
        self.word_received  = Signal(self.word_size)
        self.word_to_send   = Signal.like(self.word_received)
        self.word_complete  = Signal()

        # Status
        self.idle    = Signal()
        self.stalled = Signal()


    def elaborate(self, platform):

        m = Module()
        spi = self.spi

        # Detect falling edge of SPI clock
        past_sck    = Signal()
        m.d.sync   += past_sck.eq(spi.sck)
        sample_edge = past_sck & ~spi.sck

        # Bit counter: counts the number of bits received.
        max_bit_count = max(self.word_size, self.command_size)
        bit_count = Signal(range(0, max_bit_count + 1))

        # Shift registers for our command and data.
        current_command = Signal.like(self.command)
        current_word    = Signal.like(self.word_received)

        # De-assert our control signals unless explicitly asserted.
        m.d.sync += [
            self.command_ready.eq(0),
            self.word_complete.eq(0)
        ]


        with m.FSM() as fsm:
            m.d.comb += [
                self.idle     .eq(fsm.ongoing('IDLE')),
                self.stalled  .eq(fsm.ongoing('STALL')),
            ]

            # STALL: entered when we can't accept new bits -- either when
            # CS starts asserted, or when we've received more data than expected.
            with m.State("STALL"):

                # Wait for CS to clear.
                with m.If(~spi.cs):
                    m.next = 'IDLE'


            # We ignore all data until chip select is asserted, as that data Isn't For Us (TM).
            # We'll spin and do nothing until the bus-master addresses us.
            with m.State('IDLE'):
                m.d.sync += bit_count.eq(0)

                with m.If(spi.cs):
                    m.next = 'RECEIVE_COMMAND'


            # Once CS is low, we'll shift in our command.
            with m.State('RECEIVE_COMMAND'):

                # If CS is de-asserted early; our transaction is being aborted.
                with m.If(~spi.cs):
                    m.next = 'IDLE'

                # Continue shifting in data until we have a full command.
                with m.If(bit_count < self.command_size):
                    with m.If(sample_edge):
                        m.d.sync += [
                            bit_count       .eq(bit_count + 1),
                            current_command .eq(Cat(spi.sdi, current_command[:-1]))
                        ]

                # ... and then pass that command out to our controller.
                with m.Else():
                    m.d.sync += [
                        bit_count          .eq(0),
                        self.command_ready .eq(1),
                        self.command       .eq(current_command)
                    ]
                    m.next = 'PROCESSING'


            # Give our controller a wait state to prepare any response they might want to...
            with m.State('PROCESSING'):
                m.next = 'LATCH_OUTPUT'


            # ... and then latch in the response to transmit.
            with m.State('LATCH_OUTPUT'):
                m.d.sync += current_word.eq(self.word_to_send)
                m.next = 'SHIFT_DATA'


            # Finally, exchange data.
            with m.State('SHIFT_DATA'):

                # If CS is de-asserted early; our transaction is being aborted.
                with m.If(~spi.cs):
                    m.next = 'IDLE'

                m.d.sync += spi.sdo.eq(current_word[-1])

                # Continue shifting data until we have a full word.
                with m.If(bit_count < self.word_size):
                    with m.If(sample_edge):
                        m.d.sync += [
                            bit_count    .eq(bit_count + 1),
                            current_word .eq(Cat(spi.sdi, current_word[:-1]))
                        ]

                # ... and then output that word on our bus.
                with m.Else():
                    m.d.sync += [
                        bit_count          .eq(0),
                        self.word_complete .eq(1),
                        self.word_received .eq(current_word)
                    ]

                    # Stay in the stall state until CS is de-asserted.
                    m.next = 'STALL'

        return m


class SPIRegisterInterface(Elaboratable):
    """ SPI device interface that allows for register reads and writes via SPI.
    The SPI transaction format matches:

        in:  WAAAAAAA[...] VVVVVVVV[...]
        out: XXXXXXXX[...] RRRRRRRR[...]

    Where:
        W = write bit; a '1' indicates that the provided value is a write request
        A = all bits of the address
        V = value to be written into the register, if W is set
        R = value to be read from the register

    I/O signals:
        I: sck           -- SPI clock, from the SPI master
        I: sdi           -- SPI data in
        O: sdo           -- SPI data out
        I: cs            -- chip select, active high (as we assume your I/O will use PinsN)

        O: idle          -- true iff the register interface is currently doing nothing
        O: stalled       -- true iff the register interface cannot accept data until this transaction ends

    Other I/O ports are added dynamically with add_register().
    """

    def __init__(self, address_size=15, register_size=32, default_read_value=0, support_size_autonegotiation=True):
        """
        Parameters:
            address_size       -- the size of an address, in bits; recommended to be one bit
                                  less than a binary number, as the write command is formed by adding a one-bit
                                  write flag to the start of every address
            register_size      -- The size of any given register, in bits.
            default_read_value -- The read value read from a non-existent or write-only register.

            support_size_autonegotiation --
                If set, register 0 is used as a size auto-negotiation register. Functionally equivalent to
                calling .support_size_autonegotiation(); see its documentation for details on autonegotiation.
        """

        self.address_size  = address_size
        self.register_size = register_size
        self.default_read_value  = default_read_value

        #
        # I/O port
        #
        self.idle    = Signal()
        self.stalled = Signal()

        # Create our SPI I/O.
        self.spi = SPIBus()

        #
        # Internal details.
        #

        # Instantiate an SPI command transciever submodule.
        self.interface = SPICommandInterface(command_size=address_size + 1, word_size=register_size)

        # Create a new, empty dictionary mapping registers to their signals.
        self.registers = {}

        # Create signals for each of our register control signals.
        self._is_write = Signal()
        self._address  = Signal(self.address_size)

        if support_size_autonegotiation:
            self.support_size_autonegotiation()


    def _ensure_register_is_unused(self, address):
        """ Checks to make sure a register address isn't in use before issuing it. """

        if address in self.registers:
            raise ValueError("can't add more than one register with address 0x{:x}!".format(address))


    def support_size_autonegotiation(self):
        """ Support autonegotiation of register and address size. Consumes address 0.

        Auto-negotiation of size is relatively simple: the host sends a string of zeroes over
        the SPI bus, and we respond with:

            -- as many zeroes as there are address bits
            -- as many ones as there are data bits
            -- zeroes for any bits after

        In practice, this is functionally identical to setting register zero to a constant of all 1's.
        """
        self.add_read_only_register(0, read=-1)


    def add_sfr(self, address, *, read=None, write_signal=None, write_strobe=None, read_strobe=None):
        """ Adds a special function register to the given command interface.

        Parameters:
            address       -- the register's address, as a big-endian integer
            read          -- a Signal or integer constant representing the
                             value to be read at the given address; if not provided, the default
                             value will be read
            read_strobe   -- a Signal that is asserted when a read is completed; if not provided,
                             the relevant strobe will be left unconnected
            write_signal  -- a Signal set to the value to be written when a write is requested;
                             if not provided, writes will be ignored
            wrote_strobe  -- a Signal that goes high when a value is available for a write request
         """

        assert address < (2 ** self.address_size)
        self._ensure_register_is_unused(address)

        # Add the register to our collection.
        self.registers[address] = {
            'read': read,
            'write_signal': write_signal,
            'write_strobe': write_strobe,
            'read_strobe': read_strobe,
            'elaborate': None,
        }


    def add_read_only_register(self, address, *, read, read_strobe=None):
        """ Adds a read-only register.

        Parameters:
            address       -- the register's address, as a big-endian integer
            read          -- a Signal or integer constant representing the
                             value to be read at the given address; if not provided, the default
                             value will be read
            read_strobe   -- a Signal that is asserted when a read is completed; if not provided,
                             the relevant strobe will be left unconnected
        """
        self.add_sfr(address, read=read, read_strobe=read_strobe)



    def add_register(self, address, *, value_signal=None, size=None, name=None, read_strobe=None,
        write_strobe=None, init=0):
        """ Adds a standard, memory-backed register.

            Parameters:
                address       -- the register's address, as a big-endian integer
                value_signal  -- the signal that will store the register's value; if omitted
                                 a storage register will be created automatically
                size          -- if value_signal isn't provided, this sets the size of the created register
                reset         -- if value_signal isn't provided, this sets the reset value of the created register
                read_strobe   -- a Signal to be asserted when the register is read; ignored if not provided
                write_strobe  -- a Signal to be asserted when the register is written; ignored if not provided

            Returns:
                value_signal  -- a signal that stores the register's value; which may be the value_signal arg,
                                 or may be a signal created during execution
        """
        self._ensure_register_is_unused(address)

        # Generate a name for the register, if we don't already have one.
        name = name if name else "register_{:x}".format(address)

        # Generate a backing store for the register, if we don't already have one.
        if value_signal is None:
            size = self.register_size if (size is None) else size
            value_signal = Signal(size, name=name, init=init)

        # If we don't have a write strobe signal, create an internal one.
        if write_strobe is None:
            write_strobe = Signal(name=name + "_write_strobe")

        # Create our register-value-input and our write strobe.
        write_value  = Signal.like(value_signal, name=name + "_write_value")

        # Create a generator for a the fragments that will manage the register's memory.
        def _elaborate_memory_register(m):
            with m.If(write_strobe):
                m.d.sync += value_signal.eq(write_value)

        # Add the register to our collection.
        self.registers[address] = {
            'read': value_signal,
            'write_signal': write_value,
            'write_strobe': write_strobe,
            'read_strobe': read_strobe,
            'elaborate': _elaborate_memory_register,
        }

        return value_signal


    def _elaborate_register(self, m, register_address, connections):
        """ Generates the hardware connections that handle a given register. """

        #
        # Elaborate our register hardware.
        #

        # Create a signal that goes high iff the given register is selected.
        register_selected = Signal(name="register_address_matches_{:x}".format(register_address))
        m.d.comb += register_selected.eq(self._address == register_address)

        # Our write signal is always connected to word_received; but it's only meaningful
        # when write_strobe is asserted.
        if connections['write_signal'] is not None:
            m.d.comb += connections['write_signal'].eq(self.interface.word_received)

        # If we have a write strobe, assert it iff:
        #  - this register is selected
        #  - the relevant command is a write command
        #  - we've just finished receiving the command's argument
        if connections['write_strobe'] is not None:
            m.d.comb += [
                connections['write_strobe'].eq(self._is_write & self.interface.word_complete & register_selected)
            ]

        # Create essentially the same connection with the read strobe.
        if connections['read_strobe'] is not None:
            m.d.comb += [
                connections['read_strobe'].eq(~self._is_write & self.interface.word_complete & register_selected)
            ]

        # If we have any additional code that assists in elaborating this register, run it.
        if connections['elaborate']:
            connections['elaborate'](m)


    def _connect_interface(self, m):
        """ Connects up our SPI transciever interface.

        Intended to be overridden, if need be.
        """

        # Connect up our SPI transceiver submodule.
        m.d.comb += [
            self.interface.spi  .connect(self.spi),
            self.idle           .eq(self.interface.idle),
            self.stalled        .eq(self.interface.stalled)
        ]


    def elaborate(self, platform):
        m = Module()

        # Attach our SPI interface.
        m.submodules.interface = self.interface
        self._connect_interface(m)

        # Split the command into our "write" and "address" signals.
        m.d.comb += [
            self._is_write.eq(self.interface.command[-1]),
            self._address .eq(self.interface.command[0:-1])
        ]

        # Create the control/write logic for each of our registers.
        for address, connections in self.registers.items():
            self._elaborate_register(m, address, connections)


        # Build the logic to select the 'to_send' value, which is selected
        # from all of our registers according to the selected register address.
        first_item = True
        for address, connections in self.registers.items():
            statement = m.If if first_item else m.Elif

            with statement(self._address == address):

                # Hook up the word-to-send signal either to the read value for the relevant
                # register, or to the default read value.
                if connections['read'] is not None:
                    m.d.comb += self.interface.word_to_send.eq(connections['read'])
                else:
                    m.d.comb += self.interface.word_to_send.eq(self.default_read_value)

            # We've already created the m.If; from now on, use m.Elif
            first_item = False

        # Finally, tie all non-handled register values to always respond with the default.
        with m.Else():
            m.d.comb += self.interface.word_to_send.eq(self.default_read_value)

        return m




class SPIMultiplexer(Elaboratable):
    """ Gateware that assists in connecting multiple SPI busses to the same shared lines. """

    def __init__(self, multiplexed_busses):
        """
        Parameters:
            multiplexed_busses -- A list of SPI busses to be multiplexed. The active bus will
                                  be selected based on each bus's chip-select signals.
        """
        self.multiplexed_busses = multiplexed_busses

        #
        # I/O port
        #
        self.shared_lines = SPIBus()


    def elaborate(self, platform):
        m = Module()

        # Build our multiplexing logic.
        is_first_entry = True
        for bus in self.multiplexed_busses:
            conditional = m.If if is_first_entry else m.Elif
            is_first_entry = False

            # Connect SDO only if this line is selected
            with conditional(bus.cs):
                m.d.comb += self.shared_lines.sdo.eq(bus.sdo)


        # Connect each of our shared inputs.
        for bus in self.multiplexed_busses:
            m.d.comb += [
                bus.sck .eq(self.shared_lines.sck),
                bus.sdi .eq(self.shared_lines.sdi)
            ]


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Integrated logic analysis helpers. """

import io
import os
import sys
import math
import tempfile
import subprocess

from abc                 import ABCMeta, abstractmethod

from amaranth            import Cat, DomainRenamer, Elaboratable, Module, Signal
from amaranth.lib.cdc    import FFSynchronizer
from amaranth.lib.fifo   import AsyncFIFOBuffered
from amaranth.lib.memory import Memory
from vcd                 import VCDWriter
from vcd.gtkw            import GTKWSave

from ..stream            import StreamInterface
from ..interface.uart    import UARTMultibyteTransmitter
from ..interface.spi     import SPIDeviceInterface, SPIBus


class IntegratedLogicAnalyzer(Elaboratable):
    """ Super-simple integrated-logic-analyzer generator class for LUNA.

    Attributes
    ----------
    trigger: Signal(), input
        A strobe that determines when we should start sampling.
    sampling: Signal(), output
        Indicates when sampling is in progress.

    complete: Signal(), output
        Indicates when sampling is complete and ready to be read.

    captured_sample_number: Signal(), input
        Selects which sample the ILA will output. Effectively the address for the ILA's
        sample buffer.
    captured_sample: Signal(), output
        The sample corresponding to the relevant sample number.
        Can be broken apart by using Cat(*signals).

    Parameters
    ----------
    signals: iterable of Signals
        An iterable of signals that should be captured by the ILA.
    sample_depth: int
        The depth of the desired buffer, in samples.

    domain: string
        The clock domain in which the ILA should operate.
    sample_rate: float
        Cosmetic indication of the sample rate. Used to format output.
    samples_pretrigger: int
        The number of our samples which should be captured _before_ the trigger.
        This also can act like an implicit synchronizer; so asynchronous inputs
        are allowed if this number is >= 2. Note that the trigger strobe is read
        on the rising edge of the clock.
    """

    def __init__(self, *, signals, sample_depth, domain="sync", sample_rate=60e6, samples_pretrigger=1):
        self.domain             = domain
        self.signals            = signals
        self.inputs             = Cat(*signals)
        self.sample_width       = len(self.inputs)
        self.sample_depth       = sample_depth
        self.samples_pretrigger = samples_pretrigger
        self.sample_rate        = sample_rate
        self.sample_period      = 1 / sample_rate

        #
        # Create a backing store for our samples.
        #
        self.mem = Memory(shape=self.sample_width, depth=sample_depth, init=[])


        #
        # I/O port
        #
        self.trigger  = Signal()
        self.sampling = Signal()
        self.complete = Signal()

        self.captured_sample_number = Signal(range(0, self.sample_depth))
        self.captured_sample        = Signal(self.sample_width)


    def elaborate(self, platform):
        m  = Module()

        # Memory ports.
        write_port = self.mem.write_port()
        read_port  = self.mem.read_port(domain="sync")
        m.submodules['ila_buffer'] = self.mem

        # If necessary, create synchronized versions of the relevant signals.
        if self.samples_pretrigger >= 2:
            delayed_inputs = Signal.like(self.inputs)
            m.submodules += FFSynchronizer(self.inputs,  delayed_inputs,
                stages=self.samples_pretrigger)
        elif self.samples_pretrigger == 1:
            delayed_inputs = Signal.like(self.inputs)
            m.d.sync += delayed_inputs.eq(self.inputs)
        else:
            delayed_inputs  = self.inputs

        # Counter that keeps track of our write position.
        write_position = Signal(range(0, self.sample_depth))

        # Set up our write port to capture the input signals,
        # and our read port to provide the output.
        m.d.comb += [
            write_port.data        .eq(delayed_inputs),
            write_port.addr        .eq(write_position),

            self.captured_sample   .eq(read_port.data),
            read_port.addr         .eq(self.captured_sample_number)
        ]

        # Don't sample unless our FSM asserts our sample signal explicitly.
        m.d.sync += write_port.en.eq(0)

        with m.FSM(name="ila_state") as fsm:

            m.d.comb += self.sampling.eq(~fsm.ongoing("IDLE"))

            # IDLE: wait for the trigger strobe
            with m.State('IDLE'):

                with m.If(self.trigger):
                    m.next = 'SAMPLE'

                    # Grab a sample as our trigger is asserted.
                    m.d.sync += [
                        write_port.en  .eq(1),
                        write_position .eq(0),

                        self.complete  .eq(0),
                    ]

            # SAMPLE: do our sampling
            with m.State('SAMPLE'):

                # Sample until we run out of samples.
                m.d.sync += [
                    write_port.en  .eq(1),
                    write_position .eq(write_position + 1),
                ]

                # If this is the last sample, we're done. Finish up.
                with m.If(write_position + 1 == self.sample_depth):
                    m.next = "IDLE"

                    m.d.sync += [
                        self.complete .eq(1),
                        write_port.en .eq(0)
                    ]


        # Convert our sync domain to the domain requested by the user, if necessary.
        if self.domain != "sync":
            m = DomainRenamer(self.domain)(m)

        return m


class SyncSerialILA(Elaboratable):
    """ Super-simple ILA that reads samples out over a simple unidirectional SPI.
    Create a receiver for this object by calling apollo_fpga.ila_receiver_for(<this>).

    This protocol is simple: every time CS goes low, we begin sending out a bit of
    sample on each rising edge. Once a new sample is complete, the next sample begins
    on the next 32-bit boundary.

    Attributes
    ----------
    trigger: Signal(), input
        A strobe that determines when we should start sampling.
    sampling: Signal(), output
        Indicates when sampling is in progress.
    complete: Signal(), output
        Indicates when sampling is complete and ready to be read.

    sck: Signal(), input
        Serial clock for the SPI lines.
    sdo: Signal(), output
        Serial data out for the SPI lines.
    cs: Signal(), input
        Chip select for the SPI lines.

    Parameters
    ----------
    signals: iterable of Signals
        An iterable of signals that should be captured by the ILA.
    sample_depth: int
        The depth of the desired buffer, in samples.

    domain: string
        The clock domain in which the ILA should operate.
    samples_pretrigger: int
        The number of our samples which should be captured _before_ the trigger.
                                  This also can act like an implicit synchronizer; so asynchronous inputs
                                  are allowed if this number is >= 2.

    clock_polarity: int, 0 or 1
        Clock polarity for the output SPI transciever. Optional.
    clock_phase: int, 0 or 1
        Clock phase for the output SPI transciever. Optional.
    cs_idles_high: bool, optional
        If True, the CS line will be assumed to be asserted when cs=0.
        If False or not provided, the CS line will be assumed to be asserted when cs=1.
        This can be used to share a simple two-device SPI bus, so two internal endpoints
        can use the same CS line, with two opposite polarities.
    """

    def __init__(self, *, signals, sample_depth, clock_polarity=0, clock_phase=1, cs_idles_high=False, **kwargs):

        #
        # I/O port
        #
        self.spi = SPIBus()

        #
        # Init
        #

        self.clock_phase = clock_phase
        self.clock_polarity = clock_polarity

        # Extract the domain from our keyword arguments, and then translate it to sync
        # before we pass it back below. We'll use a DomainRenamer at the boundary to
        # handle non-sync domains.
        self.domain = kwargs.get('domain', 'sync')
        kwargs['domain'] = 'sync'

        # Create our core integrated logic analyzer.
        self.ila = IntegratedLogicAnalyzer(
            signals=signals,
            sample_depth=sample_depth,
            **kwargs)

        # Copy some core parameters from our inner ILA.
        self.signals       = signals
        self.sample_width  = self.ila.sample_width
        self.sample_depth  = self.ila.sample_depth
        self.sample_rate   = self.ila.sample_rate
        self.sample_period = self.ila.sample_period

        # Figure out how many bytes we'll send per sample.
        # We'll always send things squished into 32-bit chunks, as this is what the SPI engine
        # on our Debug Controller likes most.
        words_per_sample = (self.ila.sample_width + 31) // 32

        # Bolster our bits_per_word up to a power of two...
        self.bits_per_word = words_per_sample * 4 * 8
        self.bits_per_word = 2 ** ((self.bits_per_word - 1).bit_length())

        # ... and compute how many bits should be used.
        self.bytes_per_sample = self.bits_per_word // 8

        # Expose our ILA's trigger and status ports directly.
        self.trigger  = self.ila.trigger
        self.sampling = self.ila.sampling
        self.complete = self.ila.complete


    def elaborate(self, platform):
        m  = Module()
        m.submodules.ila = self.ila

        # Keep track of the rising edge of the CS signal
        past_spi_cs       = Signal()
        m.d.sync         += past_spi_cs.eq(self.spi.cs)
        transaction_start = ~past_spi_cs & self.spi.cs 

        # Connect up our SPI transciever to our public interface.
        interface = SPIDeviceInterface(
            word_size=self.bits_per_word,
            clock_polarity=self.clock_polarity,
            clock_phase=self.clock_phase
        )
        m.submodules.spi = interface
        m.d.comb += [
            interface.spi      .connect(self.spi),

            # Always output the captured sample.
            interface.word_out .eq(self.ila.captured_sample)
        ]

        # Count where we are in the current transmission.
        current_sample_number = Signal(range(0, self.ila.sample_depth))

        # Our first piece of data is latched in when the transaction
        # starts, so we'll move on to sample #1.
        with m.If(self.spi.cs):
            with m.If(transaction_start):
                m.d.sync += current_sample_number.eq(1)

            # From then on, we'll move to the next sample whenever we're finished
            # scanning out a word (and thus our current samples are latched in).
            # This only works correctly because the SPI interface will accept a word
            # more than one clock cycle after the edge which latches the new address
            # to read, allowing the backing memory to have a clock to latch in the
            # correct value.
            with m.Elif(interface.word_accepted):
                m.d.sync += current_sample_number.eq(current_sample_number + 1)

        # Whenever CS is low, we should be providing the very first sample,
        # so reset our sample counter to 0.
        with m.Else():
            m.d.sync += current_sample_number.eq(0)


        # Ensure our ILA module outputs the right sample.
        m.d.sync += [
            self.ila.captured_sample_number .eq(current_sample_number)
        ]

        # Convert our sync domain to the domain requested by the user, if necessary.
        if self.domain != "sync":
            m = DomainRenamer(self.domain)(m)

        return m



class StreamILA(Elaboratable):
    """ Super-simple ILA that outputs its samples over a Stream.
    Create a receiver for this object by calling apollo.ila_receiver_for(<this>).

    This protocol is simple: we wait for a trigger; and then broadcast our samples.
    We broadcast one buffer of samples per each subsequent trigger.

    Attributes
    ----------
    trigger: Signal(), input
        A strobe that determines when we should start sampling.
    sampling: Signal(), output
        Indicates when sampling is in progress.
    complete: Signal(), output
        Indicates when sampling is complete and ready to be read.

    stream: output stream
        Stream output for the ILA.

    Parameters
    ----------
    signals: iterable of Signals
        An iterable of signals that should be captured by the ILA.
    sample_depth: int
        The depth of the desired buffer, in samples.

    domain: string
        The clock domain in which the ILA should operate.
    o_domain: string
        The clock domain in which the output stream will be generated.
        If omitted, defaults to the same domain as the core ILA.
    samples_pretrigger: int
        The number of our samples which should be captured _before_ the trigger.
        This also can act like an implicit synchronizer; so asynchronous inputs
        are allowed if this number is >= 2.
    """

    def __init__(self, *, signals, sample_depth, o_domain=None, **kwargs):
        # Extract the domain from our keyword arguments, and then translate it to sync
        # before we pass it back below. We'll use a DomainRenamer at the boundary to
        # handle non-sync domains.
        self.domain = kwargs.get('domain', 'sync')
        kwargs['domain'] = 'sync'

        self._o_domain = o_domain if o_domain else self.domain

        # Create our core integrated logic analyzer.
        self.ila = IntegratedLogicAnalyzer(
            signals=signals,
            sample_depth=sample_depth,
            **kwargs)

        # Copy some core parameters from our inner ILA.
        self.signals       = signals
        self.sample_width  = self.ila.sample_width
        self.sample_depth  = self.ila.sample_depth
        self.sample_rate   = self.ila.sample_rate
        self.sample_period = self.ila.sample_period

        # Bolster our bits per sample "word" up to a power of two.
        self.bits_per_sample = 2 ** ((self.ila.sample_width - 1).bit_length())
        self.bytes_per_sample = (self.bits_per_sample + 7) // 8

        #
        # I/O port
        #
        self.stream  = StreamInterface(payload_width=self.bits_per_sample)
        self.trigger = Signal()


        # Expose our ILA's trigger and status ports directly.
        self.sampling = self.ila.sampling
        self.complete = self.ila.complete


    def elaborate(self, platform):
        m  = Module()
        m.submodules.ila = ila = self.ila

        if self._o_domain == self.domain:
            in_domain_stream = self.stream
        else:
            in_domain_stream = StreamInterface(payload_width=self.bits_per_sample)

        # Count where we are in the current transmission.
        current_sample_number = Signal(range(0, ila.sample_depth))

        # Always present the current sample number to our ILA, and the current
        # sample value to the UART.
        m.d.comb += [
            ila.captured_sample_number  .eq(current_sample_number),
            in_domain_stream.payload    .eq(ila.captured_sample)
        ]

        with m.FSM():

            # IDLE -- we're currently waiting for a trigger before capturing samples.
            with m.State("IDLE"):

                # Always allow triggering, as we're ready for the data.
                m.d.comb += self.ila.trigger.eq(self.trigger)

                # Once we're triggered, move onto the SAMPLING state.
                with m.If(self.trigger):
                    m.next = "SAMPLING"


            # SAMPLING -- the internal ILA is sampling; we're now waiting for it to
            # complete. This state is similar to IDLE; except we block triggers in order
            # to cleanly avoid a race condition.
            with m.State("SAMPLING"):

                # Once our ILA has finished sampling, prepare to read out our samples.
                with m.If(self.ila.complete):
                    m.d.sync += [
                        current_sample_number  .eq(0),
                        in_domain_stream.first      .eq(1)
                    ]
                    m.next = "SENDING"


            # SENDING -- we now have a valid buffer of samples to send up to the host;
            # we'll transmit them over our stream interface.
            with m.State("SENDING"):
                data_valid = Signal(init=1)
                m.d.comb += [
                    # While we're sending, we're always providing valid data to the UART.
                    in_domain_stream.valid  .eq(data_valid),

                    # Indicate when we're on the last sample.
                    in_domain_stream.last   .eq(current_sample_number == (self.sample_depth - 1))
                ]

                # Each time the UART accepts a valid word, move on to the next one.
                with m.If(in_domain_stream.ready):
                    with m.If(data_valid):
                        m.d.sync += [
                            current_sample_number   .eq(current_sample_number + 1),
                            data_valid              .eq(0),
                            in_domain_stream.first  .eq(0)
                        ]

                        # If this was the last sample, we're done! Move back to idle.
                        with m.If(in_domain_stream.last):
                            m.next = "IDLE"
                    with m.Else():
                        m.d.sync += data_valid.eq(1)


        # If we're not streaming out of the same domain we're capturing from,
        # we'll add some clock-domain crossing hardware.
        if self._o_domain != self.domain:
            in_domain_signals  = Cat(
                in_domain_stream.first,
                in_domain_stream.payload,
                in_domain_stream.last
            )
            out_domain_signals = Cat(
                self.stream.first,
                self.stream.payload,
                self.stream.last
            )

            # Create our async FIFO...
            m.submodules.cdc = fifo = AsyncFIFOBuffered(
                width=len(in_domain_signals),
                depth=16,
                w_domain="sync",
                r_domain=self._o_domain
            )

            m.d.comb += [
                # ... fill it from our in-domain stream...
                fifo.w_data             .eq(in_domain_signals),
                fifo.w_en               .eq(in_domain_stream.valid),
                in_domain_stream.ready  .eq(fifo.w_rdy),

                # ... and output it into our outupt stream.
                out_domain_signals      .eq(fifo.r_data),
                self.stream.valid       .eq(fifo.r_rdy),
                fifo.r_en               .eq(self.stream.ready)
            ]

        # Convert our sync domain to the domain requested by the user, if necessary.
        if self.domain != "sync":
            m = DomainRenamer(self.domain)(m)

        return m


class AsyncSerialILA(Elaboratable):
    """ Super-simple ILA that reads samples out over a UART connection.
    Create a receiver for this object by calling apollo_fpga.ila_receiver_for(<this>).

    This protocol is simple: we wait for a trigger; and then broadcast our samples.
    We broadcast one buffer of samples per each subsequent trigger.

    Attributes
    ----------
    trigger: Signal(), input
        A strobe that determines when we should start sampling.
    sampling: Signal(), output
        Indicates when sampling is in progress.
    complete: Signal(), output
        Indicates when sampling is complete and ready to be read.

    tx: Signal(), output
        Serial output for the ILA.

    Parameters
    ----------
    signals: iterable of Signals
        An iterable of signals that should be captured by the ILA.
    sample_depth: int
        The depth of the desired buffer, in samples.

    divisor: int
        The number of `sync` clock cycles per bit period.

    domain: string
        The clock domain in which the ILA should operate.
    samples_pretrigger: int
        The number of our samples which should be captured _before_ the trigger.
        This also can act like an implicit synchronizer; so asynchronous inputs
        are allowed if this number is >= 2.
    """

    def __init__(self, *, signals, sample_depth, divisor, **kwargs):
        self.divisor = divisor

        #
        # I/O port
        #
        self.tx      = Signal()

        # Extract the domain from our keyword arguments, and then translate it to sync
        # before we pass it back below. We'll use a DomainRenamer at the boundary to
        # handle non-sync domains.
        self.domain = kwargs.get('domain', 'sync')
        kwargs['domain'] = 'sync'

        # Create our core integrated logic analyzer.
        self.ila = StreamILA(
            signals=signals,
            sample_depth=sample_depth,
            **kwargs)

        # Copy some core parameters from our inner ILA.
        self.signals          = signals
        self.sample_width     = self.ila.sample_width
        self.sample_depth     = self.ila.sample_depth
        self.sample_rate      = self.ila.sample_rate
        self.sample_period    = self.ila.sample_period
        self.bits_per_sample  = self.ila.bits_per_sample
        self.bytes_per_sample = self.ila.bytes_per_sample

        # Expose our ILA's trigger and status ports directly.
        self.trigger  = self.ila.trigger
        self.sampling = self.ila.sampling
        self.complete = self.ila.complete


    def elaborate(self, platform):
        m  = Module()
        m.submodules.ila = ila = self.ila

        # Create our UART transmitter, and connect it to our stream interface.
        m.submodules.uart = uart = UARTMultibyteTransmitter(
            byte_width=self.bytes_per_sample,
            divisor=self.divisor
        )
        m.d.comb +=[
            uart.stream  .stream_eq(ila.stream),
            self.tx      .eq(uart.tx)
        ]


        # Convert our sync domain to the domain requested by the user, if necessary.
        if self.domain != "sync":
            m = DomainRenamer({"sync": self.domain})(m)

        return m



class ILAFrontend(metaclass=ABCMeta):
    """ Class that communicates with an ILA module and emits useful output. """

    def __init__(self, ila):
        """
        Parameters:
            ila -- The ILA object to work with.
        """
        self.ila = ila
        self.samples = None


    @abstractmethod
    def _read_samples(self):
        """ Read samples from the target ILA. Should return an iterable of samples. """


    def _parse_sample(self, raw_sample):
        """ Converts a single binary sample to a dictionary of names -> sample values. """

        position = 0
        sample   = {}

        # Split our raw, bits(0) signal into smaller slices, and associate them with their names.
        for signal in self.ila.signals:
            signal_width = len(signal)
            signal_bits  = raw_sample[position : position + signal_width]
            position += signal_width

            sample[signal.name] = signal_bits

        return sample


    def _parse_samples(self, raw_samples):
        """ Converts raw, binary samples to dictionaries of name -> sample. """
        return [self._parse_sample(sample) for sample in raw_samples]


    def refresh(self):
        """ Fetches the latest set of samples from the target ILA. """
        self.samples = self._parse_samples(self._read_samples())


    def enumerate_samples(self):
        """ Returns an iterator that returns pairs of (timestamp, sample). """

        # If we don't have any samples, fetch samples from the ILA.
        if self.samples is None:
            self.refresh()

        timestamp = 0

        # Iterate over each sample...
        for sample in self.samples:
            yield timestamp, sample

            # ... and advance the timestamp by the relevant interval.
            timestamp += self.ila.sample_period


    def print_samples(self):
        """ Simple method that prints each of our samples; for simple CLI debugging."""

        for timestamp, sample in self.enumerate_samples():
            timestamp_scaled = 1000000 * timestamp
            print(f"{timestamp_scaled:08f}us: {sample}")



    def emit_vcd(self, filename, *, gtkw_filename=None, add_clock=True):
        """ Emits a VCD file containing the ILA samples.

        Parameters:
            filename      -- The filename to write to, or '-' to write to stdout.
            gtkw_filename -- If provided, a gtkwave save file will be generated that
                             automatically displays all of the relevant signals in the
                             order provided to the ILA.
            add_clock     -- If true or not provided, adds a replica of the ILA's sample
                             clock to make change points easier to see.
        """

        # Select the file-like object we're working with.
        if filename == "-":
            stream = sys.stdout
            close_after = False
        else:
            stream = open(filename, 'w')
            close_after = True

        # Create our basic VCD.
        with VCDWriter(stream, timescale=f"1 ns", date='today') as writer:
            first_timestamp = math.inf
            last_timestamp  = 0

            signals = {}

            # If we're adding a clock...
            if add_clock:
                clock_value  = 1
                clock_signal = writer.register_var('ila', 'ila_clock', 'integer', size=1, init=clock_value ^ 1)

            # Create named values for each of our signals.
            for signal in self.ila.signals:
                signals[signal.name] = writer.register_var('ila', signal.name, 'integer', size=len(signal))

            # Dump the each of our samples into the VCD.
            clock_time = 0
            for timestamp, sample in self.enumerate_samples():
                for signal_name, signal_value in sample.items():

                    # If we're adding a clock signal, add any changes necessary since
                    # the last value-change.
                    if add_clock:
                        while clock_time < timestamp:
                            writer.change(clock_signal, clock_time / 1e-9, clock_value)

                            clock_value ^= 1
                            clock_time  += (self.ila.sample_period / 2)

                    # Register the signal change.
                    writer.change(signals[signal_name], timestamp / 1e-9, signal_value.to_int())


        # If we're generating a GTKW, delegate that to our helper function.
        if gtkw_filename:
            assert(filename != '-')
            self._emit_gtkw(gtkw_filename, filename, add_clock=add_clock)


    def _emit_gtkw(self, filename, dump_filename, *, add_clock=True):
        """ Emits a GTKWave save file to accompany a generated VCD.

        Parameters:
            filename      -- The filename to write the GTKW save to.
            dump_filename -- The filename of the VCD that should be opened with this save.
            add_clock     -- True iff a clock signal should be added to the GTKW save.
        """

        with open(filename, 'w') as f:
            gtkw = GTKWSave(f)

            # Comments / context.
            gtkw.comment("Generated by the LUNA ILA.")

            # Add a reference to the dumpfile we're working with.
            gtkw.dumpfile(dump_filename)

            # If we're adding a clock, add it to the top of the view.
            gtkw.trace('ila.ila_clock')

            # Add each of our signals to the file.
            for signal in self.ila.signals:
                gtkw.trace(f"ila.{signal.name}")


    def interactive_display(self, *, add_clock=True):
        """ Attempts to spawn a GTKWave instance to display the ILA results interactively. """

        # Hack: generate files in a way that doesn't trip macOS's fancy guards.
        try:
            vcd_filename = os.path.join(tempfile.gettempdir(), os.urandom(24).hex() + '.vcd')
            gtkw_filename = os.path.join(tempfile.gettempdir(), os.urandom(24).hex() + '.gtkw')

            self.emit_vcd(vcd_filename, gtkw_filename=gtkw_filename)
            subprocess.run(["gtkwave", "-f", vcd_filename, "-a", gtkw_filename])
        finally:
            os.remove(vcd_filename)
            os.remove(gtkw_filename)


class AsyncSerialILAFrontend(ILAFrontend):
    """ UART-based ILA transport.

    Parameters
    ------------
    port: string
        The serial port to use to connect. This is typically a path on *nix systems.
    ila: IntegratedLogicAnalyzer
        The ILA object to work with.
    """

    def __init__(self, *args, ila, **kwargs):
        import serial

        self._port = serial.Serial(*args, **kwargs)
        self._port.reset_input_buffer()

        super().__init__(ila)


    def _split_samples(self, all_samples):
        """ Returns an iterator that iterates over each sample in the raw binary of samples. """
        from apollo_fpga.support.bits import bits

        sample_width_bytes = self.ila.bytes_per_sample

        # Iterate over each sample, and yield its value as a bits object.
        for i in range(0, len(all_samples), sample_width_bytes):
            raw_sample    = all_samples[i:i + sample_width_bytes]
            sample_length = len(Cat(self.ila.signals))

            yield bits.from_bytes(raw_sample, length=sample_length, byteorder='big')


    def _read_samples(self):
        """ Reads a set of ILA samples, and returns them. """

        sample_width_bytes = self.ila.bytes_per_sample
        total_to_read      = self.ila.sample_depth * sample_width_bytes

        # Fetch all of our samples from the given device.
        all_samples = self._port.read(total_to_read)
        return list(self._split_samples(all_samples))
